# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

"""Functions for exporting mask sets."""
import json
import os
from importlib import import_module
from math import pi

import logging

from kqcircuits.chips.chip import Chip
from kqcircuits.defaults import mask_bitmap_export_layers, chip_export_layer_clusters, default_layers, \
    default_mask_parameters
from kqcircuits.elements.flip_chip_connectors.flip_chip_connector_dc import FlipChipConnectorDc
from kqcircuits.klayout_view import resolve_default_layer_info
from kqcircuits.pya_resolver import pya
from kqcircuits.util.area import get_area_and_density
from kqcircuits.util.count_instances import count_instances_in_cell
from kqcircuits.util.geometry_helper import circle_polygon
from kqcircuits.util.geometry_json_encoder import GeometryJsonEncoder
from kqcircuits.util.netlist_extraction import export_cell_netlist
from kqcircuits.util.export_helper import export_drc_report


def export_mask_set(mask_set, skip_extras=False):
    """Exports the designs, bitmap and documentation for the mask_set."""

    export_bitmaps(mask_set)
    export_designs(mask_set)
    if not skip_extras:
        export_docs(mask_set)


def export_designs(mask_set):
    """Exports .oas and .gds files of the mask_set."""
    # export mask layouts
    for mask_layout in mask_set.mask_layouts:
        export_masks_of_face(mask_set._mask_set_dir, mask_layout, mask_set)


def export_chip(chip_cell, chip_name, chip_dir, layout, export_drc, alt_netlists=None, skip_extras=False):
    """Exports a chip used in a maskset."""

    is_pcell = chip_cell.pcell_declaration() is not None

    # save data that is only available in pcell, not static cell
    if is_pcell:
        chip_class = type(chip_cell.pcell_declaration())
        chip_params = chip_cell.pcell_parameters_by_name()

    # export .oas file with pcells (requires exporting a cell one hierarchy level above chip pcell)
    dummy_cell = layout.create_cell(chip_name)
    dummy_cell.insert(pya.DCellInstArray(chip_cell.cell_index(), pya.DTrans()))
    _export_cell(chip_dir / f"{chip_name}_with_pcells.oas", dummy_cell, "all")
    dummy_cell.delete()
    static_cell = layout.cell(layout.convert_cell_to_static(chip_cell.cell_index()))

    # save the chip .oas file with all layers and only containing static cells
    save_opts = pya.SaveLayoutOptions()
    save_opts.format = "OASIS"
    save_opts.write_context_info = False  # to save all cells as static cells
    static_cell.write(str(chip_dir / f"{chip_name}.oas"), save_opts)

    if not skip_extras:
        # export netlist
        export_cell_netlist(static_cell, chip_dir / f"{chip_name}-netlist.json", chip_cell, alt_netlists)
        # calculate flip-chip bump count
        bump_count = count_instances_in_cell(chip_cell, FlipChipConnectorDc)
        # find layer areas and densities
        layer_areas_and_densities = {}
        for layer, area, density in zip(*get_area_and_density(static_cell)):
            if area != 0.0:
                layer_areas_and_densities[layer] = {"area": f"{area:.2f}", "density": f"{density * 100:.2f}"}

        # save auxiliary chip data into json-file
        chip_json = {
            "Chip class module": chip_class.__module__ if is_pcell else None,
            "Chip class name": chip_class.__name__ if is_pcell else None,
            "Chip parameters": chip_params if is_pcell else None,
            "Bump count": bump_count,
            "Layer areas and densities": layer_areas_and_densities
        }

        with open(chip_dir / (chip_name + ".json"), "w") as f:
            json.dump(chip_json, f, cls=GeometryJsonEncoder, sort_keys=True, indent=4)

        # export .gds files for EBL or laser writer
        for cluster_name, layer_cluster in chip_export_layer_clusters.items():
            # If the chip has no shapes in the main layers of the layer cluster, should not export the chip with
            # that layer cluster.
            export_layer_cluster = False
            for layer_name in layer_cluster.main_layers:
                shapes_iter = static_cell.begin_shapes_rec(layout.layer(default_layers[layer_name]))
                if not shapes_iter.at_end():
                    export_layer_cluster = True
                    break
            if export_layer_cluster:
                # To transform the exported layer cluster chip correctly (e.g. mirroring for top chip),
                # an instance of the cell is inserted to a temporary cell with the correct transformation.
                # Was not able to get this working by just using static_cell.transform_into().
                temporary_cell = layout.create_cell(chip_name)
                temporary_cell.insert(pya.DCellInstArray(static_cell.cell_index(), default_mask_parameters[
                    layer_cluster.face_id]["chip_trans"]))
                layers_to_export = {name: layout.layer(default_layers[name]) for name in layer_cluster.all_layers()}
                path = chip_dir / f"{chip_name}-{cluster_name}.gds"
                _export_cell(path, temporary_cell, layers_to_export)
                temporary_cell.delete()

        # export drc report for the chip
        if export_drc:
            export_drc_report(chip_name, chip_dir)

    # delete the static cell which was only needed for export
    if static_cell.cell_index() != chip_cell.cell_index():
        layout.delete_cell_rec(static_cell.cell_index())


def export_masks_of_face(export_dir, mask_layout, mask_set):
    """ Exports masks for layers of a single face of a mask_set.

    Args:
        export_dir: directory for the face specific subdirectories
        mask_layout: MaskLayout object for the cell and face reference
        mask_set: MaskSet object for the name and version attributes to be included in the filename
    """
    subdir_name_for_face = get_mask_layout_full_name(mask_set, mask_layout)
    export_dir_for_face = _get_directory(export_dir / str(subdir_name_for_face))
    # export .oas file with all layers
    path = export_dir_for_face / f"{get_mask_layout_full_name(mask_set, mask_layout)}.oas"
    _export_cell(path, mask_layout.top_cell, "all")
    # export .oas files for individual optical lithography layers
    for layer_name in mask_layout.mask_export_layers:
        export_mask(export_dir_for_face, layer_name, mask_layout, mask_set)

    # Find area and density for the layers defined in mask_layout.mask_export_density_layers
    layer_infos = [resolve_default_layer_info(layer_name, mask_layout.face_id)
                   for layer_name in mask_layout.mask_export_density_layers]
    area_data = get_area_and_density(mask_layout.top_cell, layer_infos)

    wafer_area = pi * mask_layout.wafer_rad ** 2  # Use circular wafer area instead of rectangular bounding boxes
    layer_areas_and_densities = {name: {
        "area": round(area, 2),
        "density": round(area / wafer_area * 100, 2)
    } for name, area, _ in zip(*area_data)}

    mask_json = {
        "layer_areas_and_densities": layer_areas_and_densities,
    }
    with open(export_dir_for_face / (subdir_name_for_face + ".json"), "w") as f:
        json.dump(mask_json, f, cls=GeometryJsonEncoder, sort_keys=True, indent=4)


def export_mask(export_dir, layer_name, mask_layout, mask_set):
    """ Exports a mask from a single layer of a single face of a mask set.

    Args:
        export_dir: directory for the files
        layer_name: name of the layer exported as a mask. The following prefixes can be used to modify the export:
           * Prefix ``-``: invert the shapes on this layer
           * Prefix ``^``: mirror the layer (left-right)
        mask_layout: MaskLayout object for the cell and face reference
        mask_set: MaskSet object for the name and version attributes to be included in the filename
    """
    invert = False
    if layer_name.startswith('-'):
        layer_name = layer_name[1:]
        invert = True
    mirror = False
    if layer_name.startswith('^'):
        layer_name = layer_name[1:]
        mirror = True

    top_cell = mask_layout.top_cell
    layout = top_cell.layout()
    layer_info = resolve_default_layer_info(layer_name, mask_layout.face_id)
    layer = layout.layer(layer_info)
    tmp_layer = layout.layer()

    if invert:
        wafer = pya.Region(top_cell.begin_shapes_rec(layer)).merged()
        disc = pya.Region([circle_polygon(mask_layout.wafer_rad).to_itype(layout.dbu)])
        layout.copy_layer(layer, tmp_layer)
        layout.clear_layer(layer)
        top_cell.shapes(layer).insert(wafer ^ disc)

    if mirror:
        wafer = pya.Region(top_cell.begin_shapes_rec(layer)).merged()
        layout.copy_layer(layer, tmp_layer)
        layout.clear_layer(layer)
        top_cell.shapes(layer).insert(wafer.transformed(pya.Trans(2, True, 0, 0)))

    layers_to_export = {layer_info.name: layer}
    path = export_dir / (get_mask_layout_full_name(mask_set, mask_layout) + f"-{layer_info.name}.oas")
    _export_cell(path, top_cell, layers_to_export)

    if invert:
        layout.clear_layer(layer)
        layout.copy_layer(tmp_layer, layer)
    layout.delete_layer(tmp_layer)


def export_docs(mask_set, filename="Mask_Documentation.md"):
    """Exports mask documentation containing mask layouts and parameters of all chips in the mask_set."""
    file_location = str(mask_set._mask_set_dir / filename)

    with open(file_location, "w+", encoding="utf-8") as f:
        f.write("# Mask Set Name: {}\n".format(mask_set.name))
        f.write("Version: {}\n".format(mask_set.version))

        for mask_layout in mask_set.mask_layouts:

            f.write("## Mask Layout {}:\n".format(mask_layout.face_id + mask_layout.extra_id))
            mask_layout_str = get_mask_layout_full_name(mask_set, mask_layout)
            f.write(f"![alt text]({mask_layout_str}/{mask_layout_str}-mask_graphical_rep.png)\n")

            f.write("### Number of Chips in Mask Layout {}\n".format(mask_layout.face_id + mask_layout.extra_id))

            f.write("| **Chip Name** | **Amount** |\n")
            f.write("| :--- | :--- |\n")
            for chip, amount in mask_layout.chip_counts.items():
                if amount > 0:
                    f.write("| **{}** | **{}** |\n".format(chip, amount))
            f.write("\n")

        f.write("___\n")

        f.write("## Chips\n")

        for name, cell in mask_set.used_chips.items():

            path = os.path.join("Chips", name, name)

            with open(mask_set._mask_set_dir / (path + ".json"), "r") as f2:
                chip_json = json.load(f2)

            f.write("### {} Chip\n".format(name))

            f.write(f"[{path}.oas]({path}.oas)\n\n")
            f.write(f"![{name} Chip Image]({path}.png)\n")
            f.write("\n")

            f.write("### Chip Parameters\n")
            f.write("| **Parameter** | **Value** |\n")
            f.write("| :--- | :--- |\n")

            cls_name = chip_json["Chip class name"]
            if cls_name is not None:  # otherwise it is a manually designed chip without class name or pcell parameters
                cls_mod = chip_json["Chip class module"]
                params_input = chip_json["Chip parameters"]
                cls = getattr(import_module(cls_mod), cls_name)
                # get defaults and update ones with input
                params = cls().pcell_params_by_name()
                params.update(params_input)
                params_schema = cls.get_schema()
                for param_name, param_declaration in params_schema.items():
                    f.write("| **{}** | {} |\n".format(
                        param_declaration.description.replace("|", "&#124;"),
                        str(params[param_name])))
            f.write("\n")

            f.write("### Other Chip Information\n")
            f.write("| | |\n")
            f.write("| :--- | :--- |\n")

            # launcher assignments
            launcher_assignments = Chip.get_launcher_assignments(cell)
            if len(launcher_assignments) > 0:
                f.write("| **Launcher assignments** |")
                for key, value in launcher_assignments.items():
                    f.write("{} = {}, ".format(key, value))
                f.write("|\n")

            # flip-chip bump count
            bump_count = chip_json["Bump count"]
            if bump_count > 0:
                f.write(f"| **Total bump count** | {bump_count} |\n")
            f.write("\n")

            # layer area and density
            f.write("#### Layer area and density\n")
            f.write("| **Layer** | **Total area (µm^2)** | **Density (%)** |\n")
            f.write("| :--- | :--- | :--- |\n")
            for layer, area_and_density in chip_json["Layer areas and densities"].items():
                f.write(f"| {layer} | {area_and_density['area']} | {area_and_density['density']} |\n")
            f.write("\n")

            f.write("___\n")

        f.write("## Links\n")
        for mask_layout in mask_set.mask_layouts:
            mask_layout_str = get_mask_layout_full_name(mask_set, mask_layout)
            mask_layout_path = mask_set._mask_set_dir / mask_layout_str

            f.write("### Mask Files:\n")
            for file_name in os.listdir(mask_layout_path):
                if file_name.endswith(".oas"):
                    f.write(" + [{}]({})\n".format(file_name, os.path.join(mask_layout_str, file_name)))
            f.write("\n")

            f.write("### Mask Images:\n")
            for file_name in os.listdir(mask_layout_path):
                if file_name.endswith(".png"):
                    f.write("+ [{}]({})\n".format(file_name, os.path.join(mask_layout_str, file_name)))

        f.close()


def export_bitmaps(mask_set, spec_layers=mask_bitmap_export_layers):
    """Exports bitmaps for the mask_set."""
    # pylint: disable=dangerous-default-value

    # export bitmaps for mask layouts
    for mask_layout in mask_set.mask_layouts:
        mask_layout_dir_name = get_mask_layout_full_name(mask_set, mask_layout)
        mask_layout_dir = _get_directory(mask_set._mask_set_dir / str(mask_layout_dir_name))
        filename = get_mask_layout_full_name(mask_set, mask_layout)
        view = mask_set.view
        if view:
            view.focus(mask_layout.top_cell)
            view.export_all_layers_bitmap(mask_layout_dir, mask_layout.top_cell, filename=filename)
            view.export_layers_bitmaps(mask_layout_dir, mask_layout.top_cell, filename=filename,
                                       layers_set=spec_layers, face_id=mask_layout.face_id)
    # export bitmaps for chips
    chips_dir = _get_directory(mask_set._mask_set_dir / "Chips")
    for name, cell in mask_set.used_chips.items():
        chip_dir = _get_directory(chips_dir / name)
        if view:
            view.export_all_layers_bitmap(chip_dir, cell, filename=name)
    if view:
        view.focus(mask_set.mask_layouts[0].top_cell)


def _export_cell(path, cell=None, layers_to_export=None):
    if cell is None:
        error_text = "Cannot export nil cell."
        error = ValueError(error_text)
        logging.exception(error_text, exc_info=error)
        raise error
    layout = cell.layout()
    if (layers_to_export is None) or (layers_to_export == ""):
        layers_to_export = {}

    svopt = pya.SaveLayoutOptions()
    svopt.set_format_from_filename(str(path))

    if layers_to_export == "all":
        svopt.clear_cells()
        svopt.select_all_layers()
        svopt.add_cell(cell.cell_index())
        layout.write(str(path), svopt)
    else:
        items = layers_to_export.items()
        svopt.deselect_all_layers()
        svopt.clear_cells()
        svopt.add_cell(cell.cell_index())
        for _, layer in items:
            layer_info = layout.layer_infos()[layer]
            svopt.add_layer(layer, layer_info)
        svopt.write_context_info = False
        layout.write(str(path), svopt)


def _get_directory(directory):
    if not os.path.exists(str(directory)):
        os.mkdir(str(directory))
    return directory


def get_mask_layout_full_name(mask_set, mask_layout):
    return f"{mask_set.name}_v{mask_set.version}-{mask_layout.face_id}{mask_layout.extra_id}"



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).
from kqcircuits.pya_resolver import pya


class Refpoints:
    """Helper class for extracting reference points from given layer and cell.

    Once Refpoints is initialized, it can be used similar way as dictionary, where reference point text (string) field
    is the key and reference point position (pya.DPoint) is the value.

    Refpoints is implemented such that the dictionary is extracted from given layer and cell only when it's used for the
    first time. Extracting the dictionary can be relatively time-demanding process, so this way we can speed up the
    element creation process in KQC.

    Attributes:
        layer: layer specification for source of reference points
        cell: cell containing the reference points
        trans: transform for converting reference points into target coordinate system
        rec_levels: recursion level when looking for reference points from subcells. Set to 0 to disable recursion.
    """
    def __init__(self, layer, cell, trans, rec_levels):
        self.layer = layer
        self.cell = cell
        self.trans = trans
        self.rec_levels = rec_levels
        self.refpoints = None

    def dict(self):
        """Extracts and returns reference points as dictionary, where text is the key and position is the value."""
        if self.refpoints is None:
            self.refpoints = {}
            shapes_iter = pya.RecursiveShapeIterator(self.cell.layout(), self.cell, self.layer)
            if self.rec_levels is not None:
                shapes_iter.max_depth = self.rec_levels
            while not shapes_iter.at_end():
                shape = shapes_iter.shape()
                if shape.type() in (pya.Shape.TText, pya.Shape.TTextRef):
                    self.refpoints[shape.text_string] = self.trans * (shapes_iter.dtrans()*pya.DPoint(shape.text_dpos))
                shapes_iter.next()
        return self.refpoints

    def __iter__(self):
        """Returns iterator"""
        return iter(self.dict())

    def __getitem__(self, item):
        """The [] operator to return position for given reference point text."""
        return self.dict()[item]

    def __setitem__(self, item, value):
        """The [] operator to set a new reference point."""
        self.dict()[item] = value

    def items(self):
        """Returns a list of text-position pairs."""
        return self.dict().items()

    def keys(self):
        """Returns a list of texts."""
        return self.dict().keys()

    def values(self):
        """Returns a list of positions."""
        return self.dict().values()

class RefpointToSimPort:
    """Class that takes a refpoint of an Element class with given string
    and places appropriate Simulation port(s) at the refpoint's location
    if the simulation object was instantiated using the `get_single_element_sim_class`
    class builder.

    Attributes:
        refpoint: Refpoint name string
        face: index of the face where the refpoint is located
    """
    def __init__(self, refpoint, face=0):
        self.refpoint, self.face = refpoint, face

class RefpointToInternalPort(RefpointToSimPort):
    """Creates an InternalPort at refpoint with given string
    """
    def __init__(self, refpoint, ground_refpoint,
                 resistance=50, reactance=0, inductance=0, capacitance=0,
                 face=0, junction=False, signal_layer='signal'):
        super().__init__(refpoint, face)
        self.ground_refpoint, self.resistance, self.reactance, self.inductance, self.capacitance, \
            self.junction, self.signal_layer = \
            ground_refpoint, resistance, reactance, inductance, capacitance, junction, signal_layer

class RefpointToEdgePort(RefpointToSimPort):
    """Creates an EdgePort at refpoint with given string
    """
    def __init__(self, refpoint,
                 resistance=50, reactance=0, inductance=0, capacitance=0,
                 face=0, deembed_len=None, junction=False):
        super().__init__(refpoint, face)
        self.resistance, self.reactance, self.inductance, self.capacitance, \
            self.deembed_len, self.junction = \
            resistance, reactance, inductance, capacitance, deembed_len, junction


class WaveguideToSimPort(RefpointToSimPort):
    """A waveguide is created leading to the port at the refpoint with given string in the Simulation object

    Attributes:
        refpoint: Refpoint name string
        face: index of the face where the `refpoint` is located
        towards: Another refpoint name string towards which direction the waveguide will extend.
                 If set to None, will default to "{refpoint}_corner"
        side: Indicate on which edge the port should be located. Ignored for internal ports.
              Must be one of `left`, `right`, `top` or `bottom`
        use_internal_ports: if True, ports will be inside the simulation. If False, ports will be
            brought out to an edge of the box, determined by `side`.
            Defaults to the value of the `use_internal_ports` parameter
        waveguide_length: length of the waveguide (μm), used only for internal ports
            Defaults to the value of the `waveguide_length` parameter
        term1: Termination gap (μm) at the location of `refpoint`
        turn_radius: Turn radius of the waveguide. Not relevant for internal ports.
            Defaults to the value of the `r` parameter
        a: Center conductor width. Defaults to the value of the `a` parameter
        b: Conductor gap width. Defaults to the value of the `b` parameter
        over_etching: Expansion of gaps. Defaults to the value of the `over_etching` parameter
        airbridge: if True, an airbridge will be inserted at location of the `refpoint`. Default False
    """
    def __init__(self, refpoint, face=0, towards=None, side=None, use_internal_ports=None,
                 waveguide_length=None, term1=0, turn_radius=None, a=None, b=None, over_etching=None, airbridge=False):
        super().__init__(refpoint, face)
        self.towards, self.side, self.use_internal_ports, self.waveguide_length, \
            self.term1, self.turn_radius, self.a, self.b, self.over_etching, self.airbridge = \
            towards, side, use_internal_ports, waveguide_length, term1, turn_radius, a, b, over_etching, airbridge

class JunctionSimPort(RefpointToSimPort):
    """Creates internal ports for a junction in the Simulation object.

    Depending on the value of the `separate_island_internal_ports` parameter, will either create
    two internal ports at both ends of the junction, or one port that covers both junctions.

    Attributes:
        refpoint: Refpoint name string. Defaults to "port_squid_a" as most commonly used junction port name
        other_refpoint: Refpoint name string of the other end of the junction
                        Defaults to "port_squid_b" as most commonly used junction port name
        face: index of the face where the `refpoint` is located
    """
    def __init__(self, refpoint="port_squid_a", other_refpoint="port_squid_b", face=0):
        super().__init__(refpoint, face)
        self.other_refpoint = other_refpoint



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).
import math
from collections.abc import Sequence

from typing import Tuple
from tqdm import tqdm

from kqcircuits.pya_resolver import pya
from kqcircuits.defaults import default_layers, default_brand, default_faces, default_mask_parameters, \
    default_layers_to_mask, default_covered_region_excluded_layers, default_mask_export_layers, default_bar_format
from kqcircuits.elements.markers.marker import Marker
from kqcircuits.elements.markers.mask_marker_fc import MaskMarkerFc
from kqcircuits.util.geometry_helper import circle_polygon
from kqcircuits.util.label import produce_label, LabelOrigin
from kqcircuits.util.merge import merge_layout_layers_on_face, convert_child_instances_to_static


class MaskLayout:
    """Class representing the mask for a certain face.

    A MaskLayout is used to create the cells for the mask.

    Attributes:
        layout: pya.Layout for this mask
        name: Name of the mask
        version: Mask version
        with_grid: Boolean determining if ground grid is generated
        face_id: face_id of this mask layout, "1t1" | "2b1" | "2t1"
        layers_to_mask: dictionary of layers with mask label postfix for mask label and mask covered region creation
        covered_region_excluded_layers: list of layers in `layers_to_mask` for which mask covered region is not created
        chips_map: List of lists (2D-array) of strings, each string is a chip name (or --- for no chip)
        align_to: optional exact point of placement, an (x,  y) coordinate tuple. By default the mask is centered.
        chips_map_legend: Dictionary where keys are chip names, values are chip cells
        wafer_rad: Wafer radius
        wafer_center: Wafer center as a pya.DVector
        chips_map_offset: Offset to make chips_map centered on wafer
        wafer_top_flat_length: length of flat edge at the top of the wafer
        wafer_bottom_flat_length: length of flat edge at the bottom of the wafer
        dice_width: Dicing width for this mask layout
        text_margin: Text margin for this mask layout
        chip_size: side width of the chips (for square chips), or tuple (width, height) for rectangular chips
        edge_clearance: minimum clearance of outer chips from the edge of the mask
        remove_chips: if True (default), chips that violate edge_clearance or conflict with markers are removed from
            chip maps. Note that ``extra_chips`` are never removed.
        chip_box_offset: Offset (pya.DVector) from chip origin of the chip frame boxes for this face
        chip_trans: DTrans applied to all chips
        mask_name_offset: (DEPRECATED) mask name label offset from default position (DPoint)
        mask_name_scale: text scaling factor for mask name label (float)
        mask_name_box_margin: margin around the mask name that determines the box size around the name (float)
        mask_text_scale: text scaling factor for graphical representation layer (float)
        mask_markers_dict: dictionary of all markers to be placed and kwargs to determine their position (dict)
        mask_marker_offset: offset of mask markers from wafer center in horizontal and vertical directions (float)
        mask_export_layers: list of layer names (without face_ids) to be exported as individual mask `.oas` files
        mask_export_density_layers: list of layer names (without face_ids) for which we want to calculate the coverage
         density
        submasks: list of submasks, each element is a tuple (submask mask_layout, submask position)
        extra_id: extra string used to create unique name for mask layouts with the same face_id
        extra_chips: List of tuples (name, position, trans, position_label) for chips placed outside chips_map
            trans is an optional transformation to use in place of self.chip_trans
            position_label is an optional string that overrides the automatic chip position label in the mask grid
        top_cell: Top cell of this mask layout
        added_chips: List of (chip name, chip position, chip bounding box, chip dtrans, position_label)
            populated by chips added during build()
        chip_copies: Dictionary of ``{name_copy: properties}`` where ``properties`` contains the name and location data
            for each chip that was actually added to the mask.
        mirror_labels: Boolean, if True mask and chip copy labels are mirrored. Default False.
        bbox_face_ids: List of face_ids to consider when calcualting the bounding box of chips. Defaults to [face_id]
    """

    def __init__(self, layout, name, version, with_grid, chips_map, face_id, **kwargs):

        self.layout: pya.Layout = layout
        self.name = name
        self.version = version
        self.with_grid = with_grid
        self.face_id = face_id
        self.chips_map = chips_map
        self.chips_map_legend = None
        self.chip_bounding_boxes = None

        self.layers_to_mask = kwargs.get("layers_to_mask", default_layers_to_mask)
        self.covered_region_excluded_layers = kwargs.get("covered_region_excluded_layers",
                                                         default_covered_region_excluded_layers) + ["mask_edge"]
        self.wafer_rad = kwargs.get("wafer_rad", default_mask_parameters[self.face_id]["wafer_rad"])
        self.wafer_center = (pya.DVector(0, 0))
        self.chips_map_offset = kwargs.get("chips_map_offset",
                                           default_mask_parameters[self.face_id]["chips_map_offset"])
        self.wafer_top_flat_length = kwargs.get("wafer_top_flat_length", 0)
        self.wafer_bottom_flat_length = kwargs.get("wafer_bottom_flat_length", 0)
        self.dice_width = kwargs.get("dice_width", default_mask_parameters[self.face_id]["dice_width"])
        self.text_margin = kwargs.get("text_margin", default_mask_parameters[self.face_id]["text_margin"])
        self.chip_size = kwargs.get("chip_size", default_mask_parameters[self.face_id]["chip_size"])
        if isinstance(self.chip_size, Sequence):
            self.chip_width, self.chip_height = self.chip_size
        else:
            self.chip_width, self.chip_height = self.chip_size, self.chip_size
        self.edge_clearance = kwargs.get("edge_clearance", (self.chip_width + self.chip_height) / 4)
        self.remove_chips = kwargs.get("remove_chips", True)
        self.chip_box_offset = kwargs.get("chip_box_offset", default_mask_parameters[self.face_id]["chip_box_offset"])
        self.chip_trans = kwargs.get("chip_trans", default_mask_parameters[self.face_id]["chip_trans"])
        self.mask_name_offset = kwargs.get("mask_name_offset", pya.DPoint(0, 0))  # DEPRECATED
        self.mask_name_scale = kwargs.get("mask_name_scale", 1)
        self.mask_name_box_margin = kwargs.get("mask_name_box_margin", 1000)
        self.mask_text_scale = kwargs.get("mask_text_scale", default_mask_parameters[self.face_id]["mask_text_scale"])
        self.mask_markers_dict = kwargs.get("mask_markers_dict", {Marker: {}, MaskMarkerFc: {}})
        self.mask_markers_type = kwargs.get("mask_markers_type", "all")
        self.mask_marker_offset = kwargs.get("mask_marker_offset", default_mask_parameters[self.face_id][
            "mask_marker_offset"])
        self.mask_export_layers = kwargs.get("mask_export_layers", default_mask_export_layers)
        self.mask_export_density_layers = kwargs.get("mask_export_density_layers", [])
        self.submasks = kwargs.get("submasks", [])
        self.extra_id = kwargs.get("extra_id", "")
        self.extra_chips = kwargs.get("extra_chips", [])
        self.mirror_labels = kwargs.get("mirror_labels", False)
        self.bbox_face_ids = kwargs.get("bbox_face_ids", [self.face_id])

        self.top_cell = self.layout.create_cell(f"{self.name} {self.face_id}")
        self.added_chips = []

        self.align_to = kwargs.get("align_to", None)
        self.chip_counts = {}
        self.extra_chips_maps = []
        self.chip_array_to_export = []
        self.chip_copies = {}
        # For mask name the letter I stats at x=750
        self._mask_name_letter_I_offset = 750

        self._mask_name_box_bottom_y = 35000
        self._max_x = 0
        self._max_y = 0
        self._min_x = 0
        self._min_y = 0

    def add_chips_map(self, chips_map, align=None, align_to=None, chip_size=None, chip_trans=None):
        """Add additional chip maps to the main chip map.

        The specified extra chip map, a.k.a. sub-grid, will be attached to the main grid. It may use
        different chip size than the main grid. For convenience left and rigtht sub-grids will be
        rotated 90 degrees clockwise.

        Args:
            chips_map: List of lists (2D-array) of strings, each string is a chip name (or --- for no chip)
            align: to what side of the main grid this sub-grid attaches. Allowed values: top, left, right and bottom.
            align_to: optional exact point of placement. (x,  y) coordinate tuple
            chip_size: a different chip size may be used in each sub-grid
            chip_trans: chip transformation to use for chips in this sub-grid, defaults to self.chip_trans.
        """
        chip_size = self.chip_size if not chip_size else chip_size
        self.extra_chips_maps.append((chips_map, chip_size, align, align_to, chip_trans))

    def build(self, chips_map_legend):
        """Builds the cell hierarchy for this mask layout.

        Inserts cells copied from chips_map_legend to self.top_cell at positions determined by self.chips_map. The
        copied cells are modified to only have layers corresponding to self.face_id, and they are translated and/or
        mirrored correctly based on self.face_id. Also inserts cells for mask markers, mask name label, and the circular
        area covered by the mask.

        Args:
            chips_map_legend: Dictionary where keys are chip names, values are chip cells

        """
        self.chips_map_legend = {}
        self.chip_bounding_boxes = {}

        for name, cell in tqdm(chips_map_legend.items(), desc='Building cell hierarchy', bar_format=default_bar_format):
            self.chip_counts[name] = 0
            if [name in row for row in self.chips_map] or [chip[0] == name for chip in self.extra_chips]:

                # create copies of the chips, so that modifying these only affects the ones in this MaskLayout
                new_cell = self.layout.create_cell(name)
                new_cell.copy_tree(cell)

                # Find the bounding box encompassing base metal gap shapes in all in bbox_face_ids
                bboxes = [new_cell.dbbox_per_layer(self.layout.layer(default_faces[face_id]["base_metal_gap_wo_grid"]))
                          for face_id in self.bbox_face_ids]
                if not all(b.empty() for b in bboxes):
                    p1_xs, p1_ys, p2_xs, p2_ys = zip(*[(b.p1.x, b.p1.y, b.p2.x, b.p2.y)
                                                       for b in bboxes if not b.empty()])
                    self.chip_bounding_boxes[name] = pya.DBox(min(p1_xs), min(p1_ys), max(p2_xs), max(p2_ys))
                else:
                    self.chip_bounding_boxes[name] = pya.DBox()

                # remove layers belonging to another face
                for face_id, face_dictionary in default_faces.items():
                    if face_id != self.face_id:
                        for layer_info in face_dictionary.values():
                            shapes_iter = new_cell.begin_shapes_rec(self.layout.layer(layer_info))

                            # iterating shapes using shapes_iter.at_end() fails:
                            # https://www.klayout.de/forum/discussion/comment/4275
                            # solution is to use a separate buffer list to iterate
                            shapes = []
                            while not shapes_iter.at_end():
                                try:
                                    shapes.append(shapes_iter.shape())
                                    shapes_iter.next()
                                except ValueError:
                                    print("error occurs at %s at %s" % (name, face_id))

                            for shapes_to_remove in shapes:
                                shapes_to_remove.delete()

                self.chips_map_legend[name] = new_cell

        self.region_covered = self._mask_create_geometry()
        if len(self.submasks) > 0:
            self.region_covered = pya.Region()  # don't fill with metal gap layer if using submasks
            for submask_layout, submask_pos in self.submasks:
                self.top_cell.insert(
                    pya.DCellInstArray(submask_layout.top_cell.cell_index(),
                                       pya.DTrans(submask_pos - submask_layout.wafer_center + self.wafer_center))
                )

        maskextra_cell: pya.Cell = self.layout.create_cell("MaskExtra")
        marker_region = self._add_all_markers_to_mask(maskextra_cell)

        self._insert_mask_name_label(self.top_cell, default_layers["mask_graphical_rep"], 'G')
        # add chips from chips_map
        self._add_chips_from_map(self.chips_map, self.chip_size, None, self.align_to, marker_region)
        for (chips_map, chip_size, align, align_to, chip_trans) in self.extra_chips_maps:
            self._add_chips_from_map(chips_map, chip_size, align, align_to, marker_region, chip_trans)

        # add chips outside chips_map
        for name, pos, *optional in self.extra_chips:
            trans, position_label = None, None
            if optional and isinstance(optional[0], pya.DTrans):
                trans = optional[0]
                if len(optional) > 1 and isinstance(optional[1], str):
                    position_label = optional[1]
            else:
                trans = self.chip_trans
                if optional and isinstance(optional[0], str):
                    position_label = optional[0]
            if name in chips_map_legend:
                self.region_covered -= self._add_chip(name, pos, trans, position_label)[1]
                self.chip_counts[name] += 1

        self.region_covered -= marker_region
        self._mask_create_covered_region(maskextra_cell, self.region_covered, self.layers_to_mask)
        convert_child_instances_to_static(self.layout, maskextra_cell, only_elements=True, prune=True)
        merge_layout_layers_on_face(self.layout, maskextra_cell, self.face())

    def insert_chip_copy_labels(self, labels_cell, layers):
        """Inserts chip copy labels to all chips in this mask layout and its submasks

        Args:
            labels_cell: Cell to which the labels are inserted
            layers: list of layer names (without face_ids) where the labels are produced
        """

        # find labels_cell for this mask and each submask
        labels_cells = {self: labels_cell}
        for submask_layout, submask_pos in self.submasks:
            for inst in submask_layout.top_cell.each_inst():
                # workaround for getting the cell due to KLayout bug, see
                # https://www.klayout.de/forum/discussion/1191
                # TODO: replace by `inst_cell = inst.cell` once KLayout bug is fixed
                inst_cell = submask_layout.layout.cell(inst.cell_index)
                if inst_cell.name.startswith("ChipLabels"):
                    labels_cells[submask_layout] = inst_cell
                    break

        # find all unique x and y coords of chips and place them in the corresponding keys in chips_dict
        chips_dict = {}  # {(pos_x, pos_y): chip_name, chip_pos (in submask coordinates), chip_inst, mask_layout}
        xvals = set()
        yvals = set()
        for chip_name, pos, bbox, dtrans, position_label in self.added_chips:
            if not position_label:
                xvals.add(pos.x)
                yvals.add(pos.y)
            chips_dict[(pos.x, pos.y)] = chip_name, pos, bbox, dtrans, position_label, self
        for submask_layout, submask_pos in self.submasks:
            for chip_name, pos, bbox, dtrans, position_label in submask_layout.added_chips:
                pos2 = pos + submask_pos
                if not position_label:
                    xvals.add(pos2.x)
                    yvals.add(pos2.y)
                chips_dict[(pos2.x, pos2.y)] = chip_name, pos, bbox, dtrans, position_label, submask_layout
        # produce the labels such that chips with identical x-coordinate (y-coordinate) have identical number (letter)
        used_position_labels = set()
        for (x, y), (chip_name, _, bbox, dtrans, position_label, mask_layout) in chips_dict.items():
            labels_cell_2 = labels_cells[mask_layout]
            if x not in xvals or y not in yvals:
                i, j = None, None
            else:
                i = sorted(yvals, reverse=True).index(y)
                j = sorted(xvals).index(x)
            if not position_label:
                if i is None or j is None:
                    raise ValueError("No position_label override yet label was not automatically generated")
                position_label = MaskLayout.two_coordinates_to_position_label(i, j)
            if position_label in used_position_labels:
                raise ValueError(f"Duplicate use of chip position label {position_label}. "
                                 f"When using extra_chips, please make sure to only use unreserved position labels")
            used_position_labels.add(position_label)
            bbox_x1 = bbox.left if (bool(dtrans.is_mirror()) ^ bool(self.mirror_labels)) else bbox.right
            produce_label(labels_cell_2, position_label, dtrans * (pya.DPoint(bbox_x1, bbox.bottom)),
                          LabelOrigin.BOTTOMRIGHT, mask_layout.dice_width, mask_layout.text_margin,
                          [mask_layout.face()[layer] for layer in layers],
                          mask_layout.face()["ground_grid_avoidance"], mirror=self.mirror_labels)
            bbox_x2 = bbox.right if dtrans.is_mirror() else bbox.left
            mask_layout._add_chip_graphical_representation_layer(chip_name,
                                                                 dtrans * (pya.DPoint(bbox_x2, bbox.bottom)),
                                                                 position_label, bbox.width(), labels_cell_2)
            chip_box = pya.DBox(dtrans * bbox)
            self.chip_copies[position_label] = {
                'name_chip': chip_name,
                'i': i,
                'j': j,
                'x': chip_box.left,
                'y': chip_box.bottom,
                'width': chip_box.width(),
                'height': chip_box.height(),
            }

    def face(self):
        """Returns the face dictionary for this mask layout"""
        return default_faces[self.face_id]

    @staticmethod
    def two_coordinates_to_position_label(row: int, column: int) -> str:
        """Converts two integers to chip position label, e.g. (2,3) -> 'C03'"""
        if row < 0 or ord("A") + row > ord("Z"):
            raise ValueError(f"Row coordinate {row} out of bounds")
        if column < 0 or column > 99:
            raise ValueError(f"Column coordinate {column} out of bounds")
        return chr(ord("A") + row) + ("{:02d}".format(column))

    @staticmethod
    def position_label_to_two_coordinates(position_label: str) -> Tuple[int, int]:
        """Converts chip position label to two integer coordinate, e.g. 'C03' -> (2,3)"""
        row, col = ord(position_label[0]) - ord("A"), int(position_label[1:])
        if row < 0 or ord("A") + row > ord("Z"):
            raise ValueError(f"Letter part in {position_label} out of bounds")
        if col < 0 or col > 99:
            raise ValueError(f"Number part in {position_label} out of bounds")
        return row, col

    def _mask_create_geometry(self):
        y_clip = -14.5e4

        points = []
        for a in range(0, 256 + 1):
            x = math.cos(a / 128 * math.pi) * self.wafer_rad
            y = max(math.sin(a / 128 * math.pi) * self.wafer_rad, y_clip)
            if (y > 0 and (x > self.wafer_top_flat_length / 2 or x < -self.wafer_top_flat_length / 2)) or \
                    (y < 0 and (x > self.wafer_bottom_flat_length / 2 or x < -self.wafer_bottom_flat_length / 2)):
                points.append(pya.DPoint(self.wafer_center.x + x, self.wafer_center.y + y))

        region_covered = pya.Region(pya.DPolygon(points).to_itype(self.layout.dbu))
        return region_covered

    def _add_chips_from_map(self, chips_map, chip_size, align, align_to, marker_region, chip_trans=None):
        if chip_trans is None:
            chip_trans = self.chip_trans
        if isinstance(chip_size, Sequence):
            chip_width, chip_height = chip_size
        else:
            chip_width, chip_height = chip_size, chip_size
        orig = pya.DVector(-self.wafer_rad, self.wafer_rad) - self.chips_map_offset
        if align_to:
            orig = pya.DVector(*align_to)
        elif align:  # autoalign to the specified side of the existing layout
            w = len(chips_map[0]) * chip_width / 2
            h = len(chips_map) * chip_height
            if align == "top":
                orig = pya.DVector(-w, h + self._max_y * self.layout.dbu)
            elif align == "bottom":
                orig = pya.DVector(-w, self._min_y * self.layout.dbu)
            elif align == "left":
                orig = pya.DVector(-h + self._min_x * self.layout.dbu, w)
            elif align == "right":
                orig = pya.DVector(self._max_x * self.layout.dbu, w)
        if align in ("left", "right"):  # rotate clockwise
            chips_map = zip(*reversed(chips_map))

        region_used = pya.Region()
        allowed_region = pya.Region([circle_polygon(self.wafer_rad - self.edge_clearance).to_itype(self.layout.dbu)]) \
            - marker_region \
            - pya.Region(pya.DBox(-self.wafer_rad, self._mask_name_box_bottom_y,
                                  self.wafer_rad, self.wafer_rad).to_itype(self.layout.dbu))
        for (i, row) in enumerate(tqdm(chips_map, desc='Adding chips to mask', bar_format=default_bar_format)):
            for (j, name) in enumerate(row):
                if name == "---":
                    continue
                position = pya.DPoint(chip_width * j, -chip_height * (i + 1)) + orig
                added_chip, region_chip = self._add_chip(name, position, chip_trans, allowed_region=allowed_region,
                                                         chip_width=chip_width)
                region_used += region_chip
                if added_chip:
                    self.chip_counts[name] += 1
        self.region_covered -= region_used
        box = region_used.bbox()
        self._min_x = min(box.p1.x, self._min_x)
        self._min_y = min(box.p1.y, self._min_y)
        self._max_x = max(box.p2.x, self._max_x)
        self._max_y = max(box.p2.y, self._max_y)

    def _add_chip(self, name, position, trans, position_label=None, allowed_region=None, chip_width=None):
        """Returns a tuple (Boolean telling if the chip was added, Region which the chip covers)."""
        if chip_width is None:
            chip_width = self.chip_width
        chip_region = pya.Region()
        if name in self.chips_map_legend.keys():
            chip_cell = self.chips_map_legend[name]
            bounding_box = self.chip_bounding_boxes[name]
            bbox_offset = chip_width - bounding_box.width()
            trans = pya.DTrans(position + pya.DVector(bbox_offset, 0) - self.chip_box_offset) * trans
            chip_region = pya.Region(pya.Box(trans * bounding_box * (1 / self.layout.dbu)))
            if self.remove_chips and allowed_region is not None and chip_region.inside(allowed_region).is_empty():
                return False, pya.Region()
            self.top_cell.insert(pya.DCellInstArray(chip_cell.cell_index(), trans))
            self.added_chips.append((name, position, bounding_box, trans, position_label))
            return True, chip_region
        return False, chip_region

    def _add_all_markers_to_mask(self, maskextra_cell):
        marker_region = pya.Region()
        for marker, m_kwargs in self.mask_markers_dict.items():
            # load values into kwargs
            m_kwargs['window'] = m_kwargs.get("window", True)
            m_kwargs['face_ids'] = m_kwargs.get("face_ids", [self.face_id])
            m_kwargs['wafer_rad'] = m_kwargs.get("wafer_rad", self.wafer_rad)
            m_kwargs['edge_clearance'] = m_kwargs.get("edge_clearance", self.edge_clearance)
            cell_marker = marker.create(self.layout, **m_kwargs)
            marker_transes = marker.get_marker_locations(cell_marker, **m_kwargs)
            for trans in marker_transes:
                inst = maskextra_cell.insert(pya.DCellInstArray(cell_marker.cell_index(), trans))
                marker_region += marker.get_marker_region(inst, **m_kwargs)
        return marker_region

    def _mask_create_covered_region(self, maskextra_cell, region_covered, layers_dict):
        dbu = self.layout.dbu

        leftmost_label_x = 1e10
        labels = []
        for layer, postfix in layers_dict.items():
            label_cell, label_trans = self._create_mask_name_label(self.face()[layer], postfix)
            if label_trans.disp.x < leftmost_label_x:
                leftmost_label_x = label_trans.disp.x
            labels.append((label_cell, label_trans))
        # align left edges of mask name labels in different layers
        for (label_cell, label_trans) in labels:
            label_trans = pya.DTrans(label_trans.rot, label_trans.is_mirror(), leftmost_label_x, label_trans.disp.y)
            inst = maskextra_cell.insert(pya.DCellInstArray(label_cell.cell_index(), label_trans))
            region_covered -= pya.Region(inst.bbox()).extents(self.mask_name_box_margin / dbu)

        circle = pya.DTrans(self.wafer_center) * pya.DPath(
            [pya.DPoint(math.cos(a / 32 * math.pi) * self.wafer_rad, math.sin(a / 32 * math.pi) * self.wafer_rad)
             for a in range(0, 64 + 1)], 100)
        maskextra_cell.shapes(self.layout.layer(default_layers["mask_graphical_rep"])).insert(circle)

        maskextra_cell.shapes(self.layout.layer(default_layers["mask_graphical_rep"])).insert(region_covered)
        # remove unwanted circle boundary and filling from `layers_to_mask` which have been excluded in
        # `covered_region_excluded_layers`
        for layer_name in layers_dict.keys():
            if layer_name not in self.covered_region_excluded_layers:
                maskextra_cell.shapes(self.layout.layer(self.face()[layer_name])).insert(region_covered)

        self.top_cell.insert(pya.DCellInstArray(maskextra_cell.cell_index(), pya.DTrans()))

    def _get_chip_name(self, search_cell):
        for chip_name, cell in self.chips_map_legend.items():
            if search_cell == cell:
                return chip_name
        return ""

    def _add_chip_graphical_representation_layer(self, chip_name, position, pos_index_name, chip_width, cell):
        chip_name_text = self.layout.create_cell("TEXT", "Basic", {
            "layer": default_layers["mask_graphical_rep"],
            "text": chip_name,
            "mag": 15000 * self.mask_text_scale / len(chip_name),
        })
        pos_index_name_text = self.layout.create_cell("TEXT", "Basic", {
            "layer": default_layers["mask_graphical_rep"],
            "text": pos_index_name,
            "mag": 4000 * self.mask_text_scale,
        })
        chip_name_trans = pya.DTrans(position + pya.DVector((chip_width - chip_name_text.dbbox().width()) / 2,
                                                            self.mask_text_scale * 750))
        cell.insert(pya.DCellInstArray(chip_name_text.cell_index(), chip_name_trans))
        pos_index_trans = pya.DTrans(position + pya.DVector((chip_width - pos_index_name_text.dbbox().width()) / 2,
                                                            self.mask_text_scale * 6000))
        cell.insert(pya.DCellInstArray(pos_index_name_text.cell_index(), pos_index_trans))

    def _insert_mask_name_label(self, cell, layer, postfix=""):
        cell_mask_name, trans = self._create_mask_name_label(layer, postfix)
        inst = cell.insert(pya.DCellInstArray(cell_mask_name.cell_index(), trans))
        return inst

    def _create_mask_name_label(self, layer, postfix=""):
        if postfix != "":
            postfix = "-" + postfix
        cell_mask_name = self.layout.create_cell("TEXT", "Basic", {
            "layer": layer,
            "text": default_brand + "-" + self.name + "v" + str(self.version) + postfix,
            "mag": self.mask_name_scale * 5000.0,
        })
        cell_mask_name_h = cell_mask_name.dbbox().height()
        cell_mask_name_w = cell_mask_name.dbbox().width()
        # height of top left corner
        cell_mask_name_y = math.sqrt(
            (self.wafer_rad - self.edge_clearance) ** 2 - (cell_mask_name_w / 2 + self.mask_name_box_margin) ** 2)
        self._mask_name_box_bottom_y = cell_mask_name_y - cell_mask_name_h - 2 * self.mask_name_box_margin
        trans = pya.DTrans(- self._mask_name_letter_I_offset - cell_mask_name_w / 2,
                           cell_mask_name_y - cell_mask_name_h - self.mask_name_box_margin)
        if self.mirror_labels:
            trans *= pya.DTrans(2, True, -2 * trans.disp.x, 0)
        return cell_mask_name, trans



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.elements.element import Element
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.defaults import default_junction_type
from kqcircuits.junctions import junction_type_choices


class Junction(Element):
    """Base class for junctions without actual produce function.

    This class can represent both code generated and manually designed junctions. Thus, any junction
    can be created using code like

        `self.add_element(Junction, junction_type="JunctionName", **parameters)`,

    where "JunctionName" is either a specific junction class name or name of a manually designed
    junction cell.
    """

    LIBRARY_NAME = "Junction Library"
    LIBRARY_DESCRIPTION = "Library for junctions."
    LIBRARY_PATH = "junctions"

    default_type = default_junction_type

    junction_type = Param(pdt.TypeString, "Junction Type", default_junction_type, choices=junction_type_choices)
    junction_width = Param(pdt.TypeDouble, "Junction width for code generated element", 0.02, unit="μm",
                           docstring="Junction width (only used for code generated element)")
    junction_parameters = Param(pdt.TypeString, "Extra Junction Parameters", "{}")
    _junction_parameters = Param(pdt.TypeString, "Previous state of *_parameters", "{}", hidden=True)

    @classmethod
    def create(cls, layout, library=None, junction_type=None, **parameters):
        """Create cell for a junction in layout."""
        return cls.create_subtype(layout, library, junction_type, **parameters)[0]

    def coerce_parameters_impl(self):
        self.sync_parameters(Junction)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from math import pi, sin, cos

from kqcircuits.elements.element import Element
from kqcircuits.pya_resolver import pya
from kqcircuits.util.geometry_helper import vector_length_and_direction
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.elements.waveguide_coplanar_straight import WaveguideCoplanarStraight


def arc(r, start, stop, n):
    """ Returns list of points of an arc

    Args:
        r: radius
        start: begin angle in radians
        stop: end angle in radians
        n: number of corners in full circle

    .. MARKERS_FOR_PNG 0,91 0,100
    """
    n_steps = max(round(abs(stop - start) * n / (2 * pi)), 1)
    step = (stop - start) / n_steps
    r_corner = r / cos(step / 2)

    pts = [pya.DPoint(r * cos(start), r * sin(start))]
    for i in range(n_steps):
        alpha = start + step * (i + 0.5)
        pts.append(pya.DPoint(r_corner * cos(alpha), r_corner * sin(alpha)))
    pts.append(pya.DPoint(r * cos(stop), r * sin(stop)))
    return pts


@add_parameters_from(WaveguideCoplanarStraight, "add_metal", "ground_grid_in_trace")
class WaveguideCoplanarCurved(Element):
    """The PCell declaration of a curved segment of a coplanar waveguide.

    Coordinate origin is left at the center of the arc.
    """

    alpha = Param(pdt.TypeDouble, "Curve angle (rad)", pi)
    length = Param(pdt.TypeDouble, "Actual length", 0, unit="μm", readonly=True)

    def coerce_parameters_impl(self):
        # Update length
        self.length = self.r * abs(self.alpha)

    def build(self):

        left_inner_arc, left_outer_arc, right_inner_arc, right_outer_arc, left_protection_arc, right_protection_arc, \
            annotation_arc = WaveguideCoplanarCurved.create_curve_arcs(self, self.alpha)

        # Left gap
        pts = left_inner_arc + left_outer_arc
        shape_1 = pya.DPolygon(pts)
        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(shape_1)
        # Right gap
        pts = right_inner_arc + right_outer_arc
        shape_2 = pya.DPolygon(pts)
        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(shape_2)

        pts = annotation_arc
        shape = pya.DPath(pts, self.a)
        self.cell.shapes(self.get_layer("waveguide_path")).insert(shape)
        if self.add_metal:
            self.cell.shapes(self.get_layer("base_metal_addition")).insert(shape)

        # Protection layer
        if self.ground_grid_in_trace:
            self.add_protection(shape_1.sized(1))
            self.add_protection(shape_2.sized(1))
        else:
            # Protection layer
            pts = left_protection_arc + right_protection_arc
            self.add_protection(pya.DPolygon(pts))

    @staticmethod
    def create_curve_arcs(elem, angle):
        """Creates arcs of points for a curved waveguide.

        Args:
            elem: Element from which the waveguide parameters for the arc are taken
            angle (double): angle of the curved waveguide

        Returns:
            A tuple consisting of lists of points, each list representing one of the arcs. (left_gap_inner,
            left_gap_outer, right_gap_inner, right_gap_outer, left_protection, right_protection, annotation)
        """
        alphastart = 0
        alphastop = angle
        left_gap_inner = arc(elem.r - elem.a/2, alphastart, alphastop, elem.n)
        left_gap_outer = arc(elem.r - elem.a/2 - elem.b, alphastop, alphastart, elem.n)
        right_gap_inner = arc(elem.r + elem.a/2, alphastart, alphastop, elem.n)
        right_gap_outer = arc(elem.r + elem.a/2 + elem.b, alphastop, alphastart, elem.n)
        left_protection = arc(elem.r - elem.a/2 - elem.b - elem.margin, alphastart, alphastop, elem.n)
        right_protection = arc(elem.r + elem.a/2 + elem.b + elem.margin, alphastop, alphastart, elem.n)
        annotation = arc(elem.r, alphastart, alphastop, elem.n)
        return left_gap_inner, left_gap_outer, right_gap_inner, right_gap_outer, left_protection, right_protection,\
            annotation

    @staticmethod
    def produce_curve_termination(elem, angle, term_len, trans, face_index=0):
        """Produces termination for a curved waveguide.

        The termination consists of a rectangular polygon in the metal gap layer, and grid avoidance around it.
        The termination is placed at the position where a curved waveguide with alpha=angle and trans=trans would end.

        Args:
            elem: Element from which the waveguide parameters for the termination are taken
            angle (double): angle of the curved waveguide
            term_len (double): termination length, assumed positive
            trans (DTrans): transformation applied to the termination
            face_index (int): face index of the face in elem where the termination is created
        """
        left_inner_arc, left_outer_arc, right_inner_arc, right_outer_arc, left_protection_arc, right_protection_arc,\
            _ = WaveguideCoplanarCurved.create_curve_arcs(elem, angle)

        # direction of the termination box
        _, term_dir = vector_length_and_direction(left_outer_arc[0] - left_outer_arc[1])

        if term_len > 0:
            # metal gap for termination
            pts = [
                left_inner_arc[-1],
                left_outer_arc[0],
                left_outer_arc[0] + term_len*term_dir,
                right_outer_arc[0] + term_len*term_dir,
                right_outer_arc[0],
                right_inner_arc[-1],
            ]
            shape = pya.DPolygon(pts)
            elem.cell.shapes(elem.layout.layer(elem.face(face_index)["base_metal_gap_wo_grid"])).insert(trans*shape)

        # grid avoidance for termination
        protection_pts = [
            left_protection_arc[-1],
            left_protection_arc[-1] + (term_len + elem.margin)*term_dir,
            right_protection_arc[0] + (term_len + elem.margin)*term_dir,
            right_protection_arc[0],
        ]
        elem.add_protection(trans * pya.DPolygon(protection_pts), face_index)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).



from kqcircuits.pya_resolver import pya

from kqcircuits.chips.chip import Chip
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.elements.meander import Meander
from kqcircuits.junctions.squid import Squid
from kqcircuits.qubits.swissmon import Swissmon
from kqcircuits.elements.waveguide_coplanar import WaveguideCoplanar
from kqcircuits.elements.waveguide_coplanar_splitter import WaveguideCoplanarSplitter, t_cross_parameters
from kqcircuits.util.coupler_lib import cap_params


class Shaping(Chip):
    """The PCell declaration for a Shaping chip."""

    tunable = Param(pdt.TypeBoolean, "Tunable", False)

    def build(self):

        # Launcher
        launchers = self.produce_launchers("SMA8")

        # Finnmon
        _, finnmon_refpoints_abs = self.insert_cell(Swissmon, pya.DTrans(3, False, 4000, 5000),
            arm_width=[30, 23, 30, 23],
            arm_length=[190, 96, 160, 96],
            gap_width=[29.5, 33, 29.5, 33],
            island_r=2,
            cpl_length=[235, 0, 205],
            cpl_width=[60, 42, 60],
            cpl_gap=[110, 112, 110],
            cl_offset=[150, 150]
        )

        port_qubit_dr = finnmon_refpoints_abs["port_drive"]
        port_qubit_fl = finnmon_refpoints_abs["port_flux"]
        port_qubit_ro = finnmon_refpoints_abs["port_cplr0"]
        port_qubit_sh = finnmon_refpoints_abs["port_cplr2"]

        # Chargeline
        self.insert_cell(WaveguideCoplanar,
            path=pya.DPath([
                launchers["WN"][0],
                launchers["WN"][0] + pya.DVector(self.r, 0),
                pya.DPoint((launchers["WN"][0] + pya.DVector(self.r, 0)).x, port_qubit_dr.y),
                port_qubit_dr - pya.DVector(self.r, 0),
                port_qubit_dr
            ], 1),
            term2=self.b,
        )

        # Fluxline
        self.insert_cell(WaveguideCoplanar,
            path=pya.DPath([
                launchers["WS"][0],
                launchers["WS"][0] + pya.DVector(self.r, 0),
                pya.DPoint((launchers["WS"][0] + pya.DVector(self.r, 0)).x, port_qubit_fl.y),
                port_qubit_fl - pya.DVector(self.r, 0),
                port_qubit_fl
            ], 1),
        )

        ####### Readout resonator with the purcell filter

        segment_length_target_rr = [611.586, 1834.76, 611.586]  # from qubit to shorted end
        segment_length_target_pr = [3158.32, 789.581]  # from output to shorted end
        caps_fingers = [4, 4, 4]  # J, kappa, drive
        caps_length = [37.5, 67.9, 36.2]  # J, kappa, drive
        caps_type = ["gap", "interdigital", "gap"]  # J, kappa, drive

        # Waveguide t-cross used in multiple locations
        cross1 = self.add_element(WaveguideCoplanarSplitter, **t_cross_parameters(
            a=self.a, b=self.b, a2=self.a, b2=self.b, length_extra_side=2 * self.a, length_extra=50))
        cross1_refpoints_rel = self.get_refpoints(cross1, pya.DTrans(0, False, 0, 0))
        cross1_length = cross1_refpoints_rel["port_right"].distance(cross1_refpoints_rel["port_left"])

        # Readout resonator first segment
        wg1_end = port_qubit_ro + pya.DVector(0, segment_length_target_rr[0] - cross1_length)
        self.insert_cell(WaveguideCoplanar,
            path=pya.DPath([
                port_qubit_ro,
                port_qubit_ro + pya.DVector(0, self.r),
                wg1_end + pya.DVector(0, -self.r),
                wg1_end,
            ], 1),
        )

        waveguide_length = cross1_length + cross1_refpoints_rel["base"].distance(cross1_refpoints_rel["port_bottom"])
        _, cross1_refpoints_abs = self.insert_cell(
            cross1,
            pya.DTrans(1, False, wg1_end - pya.DTrans(1, False, 0, 0) * cross1_refpoints_rel["port_left"])
        )

        meander2_end = cross1_refpoints_abs["port_bottom"] + pya.DVector(630, 0)
        self.insert_cell(Meander,
            start=cross1_refpoints_abs["port_bottom"],
            end=meander2_end,
            length=segment_length_target_rr[1] - waveguide_length,
            meanders=2,
        )

        cross2_refpoints_rel = self.get_refpoints(cross1, pya.DTrans(2, False, 0, 0))
        port_rel_cross2_wg2 = cross2_refpoints_rel["port_right"]
        _, port_abs_cross2 = self.insert_cell(
            cross1, pya.DTrans(2, False, meander2_end - port_rel_cross2_wg2))

        # Last bit of the readout resonator
        waveguide_length = cross1_refpoints_rel["base"].distance(cross1_refpoints_rel["port_bottom"])
        self.insert_cell(WaveguideCoplanar,
            path=pya.DPath([
                port_abs_cross2["port_bottom"],
                port_abs_cross2["port_bottom"] + pya.DVector(0, (segment_length_target_rr[2] - waveguide_length))
            ], 1),
        )

        # Capacitor J
        capj = self.add_element(**cap_params(caps_fingers[0], caps_length[0], caps_type[0]))
        port_rel_capj = self.get_refpoints(capj, pya.DTrans())
        self.insert_cell(capj, pya.DTrans(port_abs_cross2["port_left"] - port_rel_capj["port_a"]))

        _, port_abs_cross3 = self.insert_cell(
            cross1, pya.DTrans(2, False, port_abs_cross2["port_left"] -
                               port_rel_capj["port_a"] + port_rel_capj["port_b"] - port_rel_cross2_wg2))
        waveguide_length = cross1_length

        meander3_end = port_abs_cross3["port_left"] + pya.DVector(900, 0)

        waveguide2 = self.add_element(WaveguideCoplanar,
            path=pya.DPath([
                meander3_end,
                meander3_end + pya.DVector(self.r, 0),
                meander3_end + pya.DVector(self.r, 400),
                meander3_end + pya.DVector(self.r, 400 + self.r),
            ], 1),
        )
        self.insert_cell(waveguide2)

        waveguide_length += waveguide2.length()

        self.insert_cell(Meander,
                         start=port_abs_cross3["port_left"],
                         end=meander3_end,
                         length=segment_length_target_pr[0] - waveguide_length,
                         meanders=3,
                         )

        # Last bit of the Purcell filter of RR
        waveguide_length = cross1_refpoints_rel["base"].distance(cross1_refpoints_rel["port_bottom"])
        wg6_end = port_abs_cross3["port_bottom"] + pya.DVector(0, (segment_length_target_pr[1] - waveguide_length))
        self.insert_cell(WaveguideCoplanar,
            path=pya.DPath([
                port_abs_cross3["port_bottom"],
                wg6_end
            ], 1),
            term2=(40 if self.tunable else 0)
        )

        # Purcell resonator SQUID
        if self.tunable:
            # SQUID refpoint at the ground plane edge
            transf = pya.DTrans(2, False, wg6_end + pya.DVector(0, 40))
            self.insert_cell(Squid, transf)

            self.insert_cell(WaveguideCoplanar,
                path=pya.DPath([
                    wg6_end + pya.DVector(-20, 40 + 15),
                    wg6_end + pya.DVector(+20 + self.r, 40 + 15),
                    pya.DPoint((wg6_end + pya.DVector(+20 + self.r, 40 + 15)).x, (launchers["NE"][0]).y - self.r),
                    launchers["NE"][0] + pya.DVector(0, -self.r),
                    launchers["NE"][0] + pya.DVector(0, 0),
                ], 1),
            )

        # Capacitor Kappa
        capk = self.add_element(**cap_params(caps_fingers[1], caps_length[1], caps_type[1]))
        port_rel_capk = self.get_refpoints(capk, pya.DTrans(1, False, 0, 0))
        _, port_abs_capk = self.insert_cell(capk, pya.DTrans(1, False,
                                           meander3_end + pya.DVector(self.r, 400 + self.r) -
                                           port_rel_capk["port_a"]))

        # Output port of the purcell resonator
        self.insert_cell(WaveguideCoplanar,
            path=pya.DPath([
                port_abs_capk["port_b"],
                port_abs_capk["port_b"] + pya.DVector(0, self.r),
                pya.DPoint((port_abs_capk["port_b"] + pya.DVector(0, self.r)).x, launchers["EN"][0].y),
                launchers["EN"][0] + pya.DVector(-self.r, 0),
                launchers["EN"][0],
            ], 1),
        )

        # Capacitor for the driveline
        capi = self.add_element(**cap_params(caps_fingers[2], caps_length[2], caps_type[2]))
        port_rel_capi = self.get_refpoints(capi, pya.DTrans(1, False, 0, 0))
        _, port_abs_capi = self.insert_cell(capi, pya.DTrans(1, False, cross1_refpoints_abs["port_right"] -
                                                             port_rel_capi["port_a"]))

        # Driveline of the readout resonator
        self.insert_cell(WaveguideCoplanar,
            path=pya.DPath([
                port_abs_capi["port_b"],
                port_abs_capi["port_b"] + pya.DVector(0, self.r),
                pya.DPoint(launchers["NW"][0].x, (port_abs_capi["port_b"] + pya.DVector(0, self.r)).y),
                launchers["NW"][0] + pya.DVector(0, -self.r),
                launchers["NW"][0] + pya.DVector(0, 0),
            ], 1),
        )

        ####### Shaping resonator with the purcell filter

        segment_length_target_rr = [634.71, 1904.13, 634.71]  # from qubit to shorted end
        segment_length_target_pr = [3253.65, 813.413]  # from output to shorted end
        caps_fingers = [4, 4, 4]  # J, kappa, drive
        caps_length = [36.8, 71.5, 36.2]  # J, kappa, drive
        caps_type = ["gap", "interdigital", "gap"]  # J, kappa, drive

        # Readout resonator first segment
        wg1_end = port_qubit_sh + pya.DVector(0, -(segment_length_target_rr[0] - cross1_length))
        self.insert_cell(WaveguideCoplanar,
            path=pya.DPath([
                port_qubit_sh,
                port_qubit_sh + pya.DVector(0, -self.r),
                wg1_end + pya.DVector(0, +self.r),
                wg1_end,
            ], 1),
        )

        waveguide_length = cross1_length + cross1_refpoints_rel["base"].distance(cross1_refpoints_rel["port_bottom"])
        _, cross1_refpoints_abs = self.insert_cell(
            cross1,
            pya.DTrans(1, False, wg1_end - pya.DTrans(1, False, 0, 0) * cross1_refpoints_rel["port_right"])
        )

        meander2_end = cross1_refpoints_abs["port_bottom"] + pya.DVector(630, 0)
        self.insert_cell(Meander,
            start=cross1_refpoints_abs["port_bottom"],
            end=meander2_end,
            length=segment_length_target_rr[1] - waveguide_length,
            meanders=2,
        )

        cross2_refpoints_rel = self.get_refpoints(cross1, pya.DTrans(0, False, 0, 0))
        port_rel_cross2_wg2 = cross2_refpoints_rel["port_left"]
        _, port_abs_cross2 = self.insert_cell(cross1,
                                                        pya.DTrans(0, False, meander2_end - port_rel_cross2_wg2))

        # Last bit of the readout resonator
        waveguide_length = cross1_refpoints_rel["base"].distance(cross1_refpoints_rel["port_bottom"])
        self.insert_cell(WaveguideCoplanar,
            path=pya.DPath([
                port_abs_cross2["port_bottom"],
                port_abs_cross2["port_bottom"] + pya.DVector(0, -(segment_length_target_rr[2] - waveguide_length))
            ], 1),
        )

        # Capacitor J
        capj = self.add_element(**cap_params(caps_fingers[0], caps_length[0], caps_type[0]))
        port_rel_capj = self.get_refpoints(capj)
        self.insert_cell(capj, pya.DTrans(port_abs_cross2["port_right"] - port_rel_capj["port_a"]))

        _, port_abs_cross3 = self.insert_cell(cross1,
                                       pya.DTrans(0, False,
                                                  port_abs_cross2["port_right"] - port_rel_capj["port_a"] +
                                                  port_rel_capj["port_b"] - port_rel_cross2_wg2))
        waveguide_length = cross1_length

        meander3_end = port_abs_cross3["port_right"] + pya.DVector(900, 0)
        self.insert_cell(WaveguideCoplanar,
            path=pya.DPath([
                meander3_end,
                meander3_end + pya.DVector(self.r, 0),
                meander3_end + pya.DVector(self.r, -400),
                meander3_end + pya.DVector(self.r, -400 - self.r),
            ], 1),
        )

        waveguide_length += waveguide2.length()

        self.insert_cell(Meander,
                         start=port_abs_cross3["port_right"],
                         end=meander3_end,
                         length=segment_length_target_pr[0] - waveguide_length,
                         meanders=3,
                         )

        # Last bit of the Purcell filter of shaping resonator
        waveguide_length = cross1_refpoints_rel["base"].distance(cross1_refpoints_rel["port_bottom"])
        wg6_end = port_abs_cross3["port_bottom"] + pya.DVector(0, -(segment_length_target_pr[1] - waveguide_length))
        self.insert_cell(WaveguideCoplanar,
            path=pya.DPath([
                port_abs_cross3["port_bottom"],
                wg6_end
            ], 1),
            term2=(40 if self.tunable else 0)
        )

        # Purcell resonator SQUID
        if self.tunable:
            # SQUID refpoint at the ground plane edge
            transf = pya.DTrans(0, False, wg6_end + pya.DVector(0, -40))
            self.insert_cell(Squid, transf)

            self.insert_cell(WaveguideCoplanar,
                path=pya.DPath([
                    wg6_end + pya.DVector(-20, -40 - 15),
                    wg6_end + pya.DVector(+20 + self.r, -40 - 15),
                    pya.DPoint((wg6_end + pya.DVector(+20 + self.r, -40 - 15)).x, (launchers["SE"][0]).y + self.r),
                    launchers["SE"][0] + pya.DVector(0, self.r),
                    launchers["SE"][0] + pya.DVector(0, 0),
                ], 1),
            )

        # Capacitor Kappa
        capk = self.add_element(**cap_params(caps_fingers[1], caps_length[1], caps_type[1]))
        port_rel_capk = self.get_refpoints(capk, pya.DTrans(3, False, 0, 0))
        _, port_abs_capk = self.insert_cell(capk, pya.DTrans(3, False,
                                                                     meander3_end + pya.DVector(self.r, -400 - self.r) -
                                                                     port_rel_capk["port_a"]))

        # Output port of the purcell resonator
        self.insert_cell(WaveguideCoplanar,
            path=pya.DPath([
                port_abs_capk["port_b"],
                port_abs_capk["port_b"] + pya.DVector(0, -self.r),
                pya.DPoint((port_abs_capk["port_b"] + pya.DVector(0, -self.r)).x, (launchers["ES"][0]).y),
                launchers["ES"][0] + pya.DVector(-self.r, 0),
                launchers["ES"][0] + pya.DVector(0, 0),
            ], 1),
        )

        # Capacitor for the driveline
        capi = self.add_element(**cap_params(caps_fingers[2], caps_length[2], caps_type[2]))
        port_rel_capi = self.get_refpoints(capi, pya.DTrans(3, False, 0, 0))
        _, port_abs_capi = self.insert_cell(
            capi, pya.DTrans(3, False, cross1_refpoints_abs["port_left"] - port_rel_capi["port_a"]))

        # Driveline of the shaping resonator
        self.insert_cell(WaveguideCoplanar,
            path=pya.DPath([
                port_abs_capi["port_b"],
                port_abs_capi["port_b"] + pya.DVector(0, -self.r),
                pya.DPoint(launchers["SW"][0].x, (port_abs_capi["port_b"] + pya.DVector(0, -self.r)).y),
                launchers["SW"][0] + pya.DVector(0, self.r),
                launchers["SW"][0] + pya.DVector(0, 0),
            ], 1),
        )



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

from sys import float_info
import textwrap

from kqcircuits.defaults import default_layers
from kqcircuits.elements.chip_frame import ChipFrame
from kqcircuits.elements.element import Element, get_refpoints, insert_cell_into
from kqcircuits.elements.waveguide_coplanar import WaveguideCoplanar
from kqcircuits.elements.waveguide_composite import WaveguideComposite, Node
from kqcircuits.util.parameters import pdt
from kqcircuits.pya_resolver import pya


def convert_cells_to_code(top_cell, print_waveguides_as_composite=False, add_instance_names=True, refpoint_snap=50.0,
                          grid_snap=1.0, output_format="insert_cell+chip", include_imports=True,
                          use_create_with_refpoints=True):

    """Prints out the Python code required to create the cells in top_cell.

    For each instance that is selected in GUI, prints out an `insert_cell()` command that can be copy pasted to a chip's
    `build()`. If no instances are selected, then it will do the same for all instances that are one level below
    the chip cell in the cell hierarchy. PCell parameters are taken into account. Waveguide points can automatically be
    snapped to closest refpoints in the generated code.

    Args:
        top_cell: cell whose child cells will be printed as code
        print_waveguides_as_composite: If true, then WaveguideCoplanar elements are printed as WaveguideComposite.
        add_instance_names: If true, then unique instance names will be added for each printed element. This is required
            if you want to have waveguides connect to refpoints of elements that were placed in GUI.
        refpoint_snap: If a waveguide point is closer than `refpoint_snap` to a refpoint, the waveguide point will be
            at that refpoint.
        grid_snap: If a waveguide point was not close enough to a refpoint, it will be snapped to a square grid with
            square side length equal to `grid_snap`
        output_format: Determines the format of the code for placing cells and if some extra code is printed. Has the
            following options:

            * "insert_cell": only insert_cell() calls which can be copied to existing chip's/element's build method
            * "insert_cell+chip": same as previous, but prints also the chip code, can copy to empty file to create
              new chip
            * "create": only create() and cell.insert() calls which can be copied to an existing macro with layout
              and top_cell
            * "create+macro": same as previous, but includes initial lines for macro, can copy to empty file to
              create new macro

        include_imports: If true, then import statements for all used elements are included in the generated code
        use_create_with_refpoints: If true, then create_with_refpoints() is used instead of create(). Only used when
            output_format is "create" or "create+macro". Required if you want to use refpoints as waveguide points.

    Returns:
        str: The generated Python code. This is also printed.
    """

    layout = top_cell.layout()

    instances = []
    inst_names = []
    pcell_classes = set()

    def add_instance(inst):
        cell = _get_cell(inst)
        inst_name = inst.property('id')
        pcell_decl = inst.pcell_declaration()
        # ChipFrame is always constructed by Chip, so we don't want to generate code for it
        if isinstance(pcell_decl, ChipFrame):
            return
        # Instance name labels are generated based on element instance names, we don't want to create them explicitly
        if cell.name == "TEXT":
            return
        instances.append(inst)
        if inst_name is not None:
            inst_names.append(inst_name)
        if pcell_decl is not None:
            pcell_classes.add(pcell_decl.__class__)

    # If some instances are selected, then we are only going to export code for them
    cell_view = pya.CellView.active() if hasattr(pya, "CellView") else None
    if cell_view and cell_view.is_valid() and len(cell_view.view().object_selection) > 0:
        for obj in cell_view.view().object_selection:
            if obj.is_cell_inst():
                add_instance(obj.inst())
    # Otherwise get all instances at one level below top_cell.
    else:
        for inst in top_cell.each_inst():
            add_instance(inst)

    # Order the instances according to cell type, x-coordinate, and y-coordinate
    instances = sorted(instances, key=lambda inst: (inst.cell.name, -inst.dtrans.disp.y, inst.dtrans.disp.x))
    pcell_classes = sorted(pcell_classes, key=lambda pcell_class: pcell_class.__module__)

    # Move all WaveguideComposite and WaveguideCoplanar elements to the end of the list. This is required so that the
    # refpoints from other instances can be used as waveguide path points. We can assume here that all
    # WaveguideComposite and WaveguideCoplanar are consecutive elements in the list due to the previous sorting
    instances = _move_to_end(instances, WaveguideComposite)
    instances = _move_to_end(instances, WaveguideCoplanar)

    # Add names to placed instances and create chip-level refpoints with those names
    if add_instance_names:
        for inst in instances:
            if inst.property("id") is None and \
                    not isinstance(inst.pcell_declaration(), (WaveguideComposite, WaveguideCoplanar)):
                inst_name = _get_unique_inst_name(inst, inst_names)
                inst_names.append(inst_name)
                inst.set_property("id", inst_name)
                inst_refpoints = get_refpoints(layout.layer(default_layers["refpoints"]), inst.cell, inst.dcplx_trans,
                                               0)
                for name, refpoint in inst_refpoints.items():
                    text = pya.DText(f"{inst_name}_{name}", refpoint.x, refpoint.y)
                    top_cell.shapes(layout.layer(default_layers["refpoints"])).insert(text)

    # Get refpoints used for snapping waveguide points
    refpoints = get_refpoints(layout.layer(default_layers["refpoints"]), top_cell)
    # only use refpoints of named instances
    refpoints = {name: point for name, point in refpoints.items() if name.startswith(tuple(inst_names))}

    # Generate code for importing the used element. More element imports may be added later when generating code from
    # waveguide nodes.
    element_imports = ""
    if include_imports:
        for pcell_class in pcell_classes:
            element_imports += f"from {pcell_class.__module__} import {pcell_class.__name__}\n"
        if print_waveguides_as_composite or WaveguideComposite in pcell_classes:
            element_imports += f"from {WaveguideComposite.__module__} import {Node.__name__}\n"

    def get_waveguide_code(inst, prefix, postfix, pcell_type, point_prefix, point_postfix="", refpoint_prefix="",
                           refpoint_postfix=""):
        path_str = prefix

        wg_points = []
        nodes = None
        _params = inst.pcell_parameters_by_name()
        if type(inst.pcell_declaration()).__name__ == "WaveguideCoplanar":
            wg_points = _params.pop("path").each_point()
        else:
            nodes = Node.nodes_from_string(_params.pop("nodes"))
            for node in nodes:
                wg_points.append(node.position)

        wg_params = ""  # non-default parameters of the cell
        for k, v in inst.pcell_declaration().get_schema().items():
            if k in _params and v.data_type != pdt.TypeShape and _params[k] != v.default:
                wg_params += f",  {k}={_params[k]}"

        for i, path_point in enumerate(wg_points):
            path_point += inst.dtrans.disp
            x_snapped = grid_snap*round(path_point.x/grid_snap)
            y_snapped = grid_snap*round(path_point.y/grid_snap)
            node_params = ""
            if nodes is not None:
                node_params, node_elem = get_node_params(nodes[i])
                if node_elem is not None and include_imports:
                    nonlocal element_imports
                    node_elem_import = f"from {node_elem.__module__} import {node_elem.__name__}\n"
                    if node_elem_import not in element_imports:
                        element_imports += node_elem_import

            # If a refpoint is close to the path point, snap the path point to it
            closest_dist = float_info.max
            closest_refpoint_name = None
            for name, point in refpoints.items():
                dist = point.distance(path_point)
                if dist <= closest_dist and dist < refpoint_snap:
                    # If this refpoint is at exact same position as closest_refpoint, compare also refpoint names.
                    # This should ensure that chip-level refpoints are chosen over lower-level refpoints.
                    if dist < closest_dist or (len(name) > len(closest_refpoint_name)):
                        closest_dist = dist
                        closest_refpoint_name = name
            if closest_refpoint_name is not None:
                if output_format.startswith("insert_cell"):
                    path_str += f"{refpoint_prefix}self.refpoints[\"{closest_refpoint_name}\"]{node_params}" \
                                f"{refpoint_postfix}, "
                elif use_create_with_refpoints:
                    refp_split = closest_refpoint_name.split("_")
                    refp_name = "_".join(refp_split[1:])
                    path_str += f"{refpoint_prefix}{refp_split[0].replace('-', '_')}_refpoints[\"{refp_name}\"]" \
                                f"{node_params}{refpoint_postfix}, "
                else:
                    path_str += f"{point_prefix}({x_snapped}, {y_snapped}){node_params}{point_postfix}, "
            else:
                path_str += f"{point_prefix}({x_snapped}, {y_snapped}){node_params}{point_postfix}, "
        path_str = path_str[:-2]  # Remove extra comma and space
        path_str += postfix + wg_params
        if output_format.startswith("insert_cell"):
            return f"self.insert_cell({pcell_type}, {path_str})\n"
        else:
            return f"view.insert_cell({pcell_type}, {path_str})\n"

    def transform_as_string(inst):
        trans = inst.dcplx_trans
        x, y = trans.disp.x, trans.disp.y
        if trans.mag == 1 and trans.angle % 90 == 0:
            if trans.rot() == 0 and not trans.is_mirror():
                if x == 0 and trans.disp.y == 0:
                    return ""
                else:
                    return f"pya.DTrans({x}, {y})"
            else:
                return f"pya.DTrans({trans.rot()}, {trans.is_mirror()}, {x}, {y})"
        else:
            return f"pya.DCplxTrans({trans.mag}, {trans.angle}, {trans.is_mirror()}, {x}, {y})"

    # Generate the code for creating each instance
    instances_code = ""
    for inst in instances:
        # Print python code for creating the instance at the given position
        cell = _get_cell(inst)
        transform = transform_as_string(inst)
        pcell_declaration = inst.pcell_declaration()
        if pcell_declaration is not None:
            if isinstance(pcell_declaration, Element):
                pcell_type = type(pcell_declaration).__name__
                # special handling for waveguides
                if isinstance(pcell_declaration, (WaveguideComposite, WaveguideCoplanar)):
                    if print_waveguides_as_composite or isinstance(pcell_declaration, WaveguideComposite):
                        instances_code += \
                            get_waveguide_code(inst, prefix="nodes=[", postfix="]", pcell_type="WaveguideComposite",
                                               point_prefix="Node(", point_postfix=")", refpoint_prefix="Node(",
                                               refpoint_postfix=")")
                    else:
                        instances_code += \
                            get_waveguide_code(inst, prefix="path=pya.DPath([", postfix="], 0)",
                                               pcell_type=pcell_type, point_prefix="pya.DPoint")
                # other elements
                else:
                    inst_name = inst.property('id') if (inst.property('id') is not None) else ""
                    var_name = inst_name.replace("-", "_")
                    transform_nonempty = transform if transform != "" else "pya.DTrans()"
                    if output_format.startswith("insert_cell"):
                        inst_name = f"inst_name=\"{inst_name}\"" if (inst_name != "") else ""
                        inst_name = f", {inst_name}" if transform != "" else inst_name
                        instances_code += \
                            f"self.insert_cell({pcell_type}, {transform}{inst_name}{_pcell_params_as_string(cell)})\n"
                    elif use_create_with_refpoints:
                        refpoint_transform = f"refpoint_transform={transform}, " if transform != "" else ""
                        instances_code += \
                            f"{var_name}, {var_name}_refpoints = {pcell_type}.create_with_refpoints(layout, " \
                            f"{refpoint_transform}rec_levels=0{_pcell_params_as_string(cell)})\n"
                        instances_code += f"view.insert_cell({var_name}, {transform_nonempty})\n"
                    else:
                        instances_code += \
                            f"{var_name} = {pcell_type}.create(layout{_pcell_params_as_string(cell)})\n"
                        instances_code += f"view.insert_cell({var_name}, {transform_nonempty})\n"
            else:
                # non-Element PCell
                if output_format.startswith("insert_cell"):
                    instances_code += \
                        f"cell = self.layout.create_cell(\"{cell.name}\", \"{cell.library().name()}\", " \
                        f"{cell.pcell_parameters_by_name()})\n"
                    instances_code += f"self.insert_cell(cell, {transform})\n"
                else:
                    instances_code += \
                        f"cell = layout.create_cell(\"{cell.name}\", \"{cell.library().name()}\", " \
                        f"{cell.pcell_parameters_by_name()})\n" \
                        f"view.insert_cell(cell, {transform})\n"

        else:
            # static cell
            if output_format.startswith("insert_cell"):
                instances_code += f"cell = self.layout.create_cell(\"{cell.name}\", \"{cell.library().name()}\")\n"
                instances_code += f"self.insert_cell(cell, {transform})\n"
            else:
                instances_code += f"cell = layout.create_cell(\"{cell.name}\", \"{cell.library().name()}\")\n"
                instances_code += f"view.insert_cell(cell, {transform})\n"

    # Generate code for the beginning of the chip or macro file if needed
    start_code = ""
    if output_format == "insert_cell+chip":
        start_code += textwrap.dedent("""\
            from kqcircuits.pya_resolver import pya
            from kqcircuits.chips.chip import Chip\n\n""")
        start_code += element_imports + "\n"
        start_code += textwrap.dedent("""\

            class NewChip(Chip):

                def build(self):\n\n""")
    elif output_format == "create+macro":
        start_code += "from kqcircuits.pya_resolver import pya\n"
        start_code += "from kqcircuits.klayout_view import KLayoutView\n\n"
        start_code += element_imports + "\n"
        start_code += "view = KLayoutView()\n"
        start_code += "layout = view.layout\n\n"
    else:
        start_code += element_imports + "\n"

    if output_format == "insert_cell+chip":
        full_code = start_code + textwrap.indent(instances_code, "        ") + "\n"
    else:
        full_code = start_code + instances_code

    return full_code


def _get_cell(inst):
    # workaround for getting the cell due to KLayout bug, see
    # https://www.klayout.de/forum/discussion/1191/cell-shapes-cannot-call-non-const-method-on-a-const-reference
    # TODO: replace by `inst.cell` once KLayout bug is fixed
    return inst.layout().cell(inst.cell_index)


def _get_unique_inst_name(inst, inst_names):
    idx = 1
    inst_name = type(inst.pcell_declaration()).__name__ + str(idx)
    while inst_name in inst_names:
        idx += 1
        inst_name = type(inst.pcell_declaration()).__name__ + str(idx)
    return inst_name


def _pcell_params_as_string(cell):
    params = cell.pcell_parameters_by_name()
    params_schema = type(cell.pcell_declaration()).get_schema()
    params_str = ""
    for param_name, param_declaration in params_schema.items():
        if (params[param_name] != param_declaration.default and param_name != "refpoints"
                and not (param_name.startswith("_") and param_name.endswith("_parameters"))):
            param_value = params[param_name]
            if isinstance(param_value, str):
                param_value = repr(param_value)
            if isinstance(param_value, pya.DPoint):
                param_value = f"pya.DPoint({param_value})"
            params_str += f", {param_name}={param_value}"
    return params_str


def _move_to_end(instances, pcell_type):
    """Returns `instances` list where instances of `pcell_type` are at the end

    Assumes that all `pcell_type` instances are consecutive elements of the list.
    """
    wg_indices = [idx for idx, inst in enumerate(instances) if isinstance(inst.pcell_declaration(), pcell_type)]
    if len(wg_indices) > 0:
        if wg_indices[-1] < len(instances) - 1:  # otherwise the waveguides are already at the end of instances list
            instances = \
                instances[:wg_indices[0]] + instances[wg_indices[-1] + 1:] + instances[wg_indices[0]:wg_indices[-1] + 1]
    return instances


def get_node_params(node: Node):
    """
    Generate a list of parameters for Node in string form

    Args:
        node: a Node to convert

    Returns: a tuple (node_params, element) where
        node_params: string of comma-separated key-value pairs that can be passed to the initializer of Node,
        starting with ``", "``
        element: class that implements the node's element, or None if the node has no element
    """
    node_params = ""
    elem = None
    for k, v in vars(node).items():
        if k == "element" and v is not None:
            node_params += f", {v.__name__}"
            elem = v
        elif (k == "inst_name" and v is not None) or \
                (k == "align" and v != tuple()) or \
                (k == "angle" and v is not None) or \
                (k == "length_before" and v is not None) or \
                (k == "length_increment" and v is not None):
            node_params += f", {k}={repr(v)}"
        elif k == "params":
            # Expand keyword arguments to Node
            for kk, vv in v.items():
                node_params += f", {kk}={repr(vv)}"
    return node_params, elem


def extract_pcell_data_from_views():
    """Remove all PCells and return their data.

    Returns: A list of lists. Each element corresponds to a view in KLayout and it is a list of
        ``(type, location, parameters)`` tuples. These tuples completely describe the type, position and
        parameters of a single PCell in the "Top Cell" of this view.
    """

    views = []
    main_window = pya.Application.instance().main_window()
    for vid in range(main_window.views()):
        top_cell = main_window.view(vid).active_cellview().cell
        pcells = []
        for inst in top_cell.each_inst():
            pc = inst.pcell_declaration()
            if pc:
                params = inst.pcell_parameters_by_name()
                def_params = pc.__class__.get_schema()
                for k, v in def_params.items():
                    if params[k] == v.default:
                        del params[k]
                pcells.append((pc.__class__, inst.dtrans, params))
                inst.delete()
        views.append(pcells)

    return views


def restore_pcells_to_views(views):
    """Re-populate each view's Top Cell with PCells as extracted by ``extract_pcell_data_from_views``.

    Args:
        views: List of list of ``(type, location, parameters)`` tuples.
    """

    main_window = pya.Application.instance().main_window()
    if main_window.views() != len(views):
        raise ValueError("Number of views in KLayout unexpectedly changed during reload.")

    for vid in range(main_window.views()):
        top_cell = main_window.view(vid).active_cellview().cell
        pcells = views[vid]
        for pc in pcells:
            def_params = {k:v.default for k, v in pc[0].get_schema().items()}
            params = {**def_params, **pc[2]}
            insert_cell_into(top_cell, pc[0], pc[1], **params)
        top_cell.refresh()



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import numpy

from kqcircuits.pya_resolver import pya


def make_grid(boundbox, avoid_region, grid_step=10, grid_size=5, group_n=10):
    """Generates ground grid covering `boundbox` with holes not overlapping with the `avoid_region`.

    Args:
        boundbox: bounding box of grid in database unit
        avoid_region: area on which grid is avoided
        grid_step: step between consecutive holes in database unit
        grid_size: hole edge length in database unit
        group_n: number of adjacent holes in a group (is used to speed up grid generation)

    Returns:
        grid region
  """
    def grid_region(box, step, size):
        square = pya.Box(0, 0, size, size)
        x_region = pya.Region()
        for x in numpy.arange(box.p1.x, box.p2.x, step):
            x_region.insert(square.transformed(pya.Trans(pya.Vector(x, 0))))
        xy_region = pya.Region()
        for y in numpy.arange(box.p1.y, box.p2.y, step):
            xy_region.insert(x_region.transformed(pya.Trans(pya.Vector(0, y))))
        return xy_region

    # Create box grid, where each box can include group_n x group_n holes.
    box_step = group_n * grid_step
    box_size = box_step - grid_step + grid_size
    boxes_region = grid_region(boundbox, box_step, box_size)
    # Filter boxes that do not overlap with avoid_region. These will be used to speed up filtering of holes.
    masked_boxes_region = ((boxes_region & pya.Region(boundbox)) - avoid_region).with_area(box_size**2, None, False)

    # Create grid of holes and duplicate it on boxes that overlap with avoid_region.
    holes_region = grid_region(pya.Box(0, 0, box_step, box_step), grid_step, grid_size)
    overlap_region = pya.Region()
    for poly in (boxes_region - masked_boxes_region).each():
        overlap_region.insert(holes_region.transformed(pya.Trans(pya.Vector(poly.bbox().p1))))

    # Create grid of holes that do not overlap with avoid region.
    masked_holes_region = ((overlap_region & pya.Region(boundbox)) - avoid_region).with_area(grid_size**2, None, False)
    for poly in masked_boxes_region.each():
        masked_holes_region.insert(holes_region.transformed(pya.Trans(pya.Vector(poly.bbox().p1))))

    return masked_holes_region



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


"""Helper module for general geometric functions"""

from math import cos, sin, radians, atan2, degrees, pi, ceil
from typing import List
import numpy as np
from scipy import spatial
from kqcircuits.defaults import default_layers, default_path_length_layers
from kqcircuits.pya_resolver import pya


def vector_length_and_direction(vector):
    """Returns the direction and length of the pya.DVector "vector"."""
    length = vector.length()
    direction = vector / length
    return length, direction


def point_shift_along_vector(start, other, distance=None):
    """Returns a point at a `distance` away from point `start` in the direction of point `other`."""
    v = other - start
    if distance is not None:
        return start + v / v.length() * distance
    else:
        return start + v


def get_direction(angle):
    """
    Returns the direction vector corresponding to `angle`.

    Args:
        angle: angle in degrees

    Returns: Unit vector in direction angle
    """
    return pya.DVector(cos(radians(angle)), sin(radians(angle)))


def get_angle(vector):
    """
    Returns the angle in degrees for a given DVector (or DPoint)

    Args:
        vector: input vector

    Returns: angle in degrees
    """
    return degrees(atan2(vector.y, vector.x))


def get_cell_path_length(cell, layer=None):
    """Returns the length of the paths in the cell.

    Adding together the cell's paths' lengths in the "1t1_waveguide_path", "2b1_waveguide_path" and
    "waveguide_length" layers.

    Args:
        cell: A cell object.
        layer: None or an unsigned int to specify a non-standard layer
    """

    if layer is not None:
        return _get_length_per_layer(cell, layer)

    length = 0
    for path_layer in default_path_length_layers:
        length += _get_length_per_layer(cell, path_layer)

    return length


def _get_length_per_layer(cell, layer):
    """Get length of the paths in the cell in the specified layer."""

    length = 0
    layer = cell.layout().layer(default_layers[layer]) if isinstance(layer, str) else layer

    for inst in cell.each_inst():  # over child cell instances, not instances of itself
        shapes_iter = inst.cell.begin_shapes_rec(layer)
        while not shapes_iter.at_end():
            shape = shapes_iter.shape()
            if shape.is_path():
                length += shape.path_dlength()
            shapes_iter.next()

    # in case of waveguide, there are no shapes in the waveguide cell itself
    # but the following allows function reuse in other applications
    for shape in cell.shapes(layer).each():
        if shape.is_path():
            length += shape.path_dlength()

    return length


def get_object_path_length(obj, layer=None):
    """Returns sum of lengths of all the paths in the object and its children

    Arguments:
        obj: ObjectInstPath object
        layer: layer integer id in the database, waveguide layer by default
    """

    if obj.is_cell_inst():
        # workaround for getting the cell due to KLayout bug, see
        # https://www.klayout.de/forum/discussion/1191/cell-shapes-cannot-call-non-const-method-on-a-const-reference
        # TODO: replace by `cell = obj.inst().cell` once KLayout bug is fixed
        cell = obj.layout().cell(obj.inst().cell_index)
        return get_cell_path_length(cell, layer)
    else:  # is a shape
        # TODO ignore paths on wrong layers
        shape = obj.shape
        if shape.is_path():
            return shape.path_dlength()
    return 0


def simple_region(region):
    return pya.Region([poly.to_simple_polygon() for poly in region.each()])


def region_with_merged_points(region, tolerance):
    """ In each polygon of the region, removes points that are closer to other points than a given tolerance.

    Arguments:
        region: Input region
        tolerance: Minimum distance, in database units, between two adjacent points in the resulting region

    Returns:
        region: with merged points
    """
    def find_next(curr, step, data):
        """ Returns the next index starting from 'i' to direction 'step' for which 'data' has positive value """
        num = len(data)
        j = curr + step
        while data[j % num] <= 0.0:
            j += step
        return j

    def merged_points(points):
        """ Removes points that are closer another points than a given tolerance. Returns list of points."""
        # find squared length of each segment of polygon
        num = len(points)
        squares = [0.0] * num
        for i in range(0, num):
            squares[i] = points[i].sq_distance(points[(i + 1) % num])

        # merge short segments
        curr_id = 0
        squared_tolerance = tolerance ** 2
        while curr_id < num:
            if squares[curr_id % num] >= squared_tolerance:
                # segment long enough: increase 'curr' for the next iteration
                curr_id = find_next(curr_id, 1, squares)
                continue

            # segment too short: merge segment with the shorter neighbor segment (prev or next)
            prev_id = find_next(curr_id, -1, squares)
            next_id = find_next(curr_id, 1, squares)
            if squares[prev_id % num] < squares[next_id % num]:  # merge with the previous segment
                squares[curr_id % num] = 0.0
                curr_id = prev_id
            else:  # merge with the next segment
                squares[next_id % num] = 0.0
                next_id = find_next(next_id, 1, squares)
            squares[curr_id % num] = points[curr_id % num].sq_distance(points[next_id % num])

        return [point for square, point in zip(squares, points) if square > 0.0]

    # Quick exit if tolerance is not positive
    if tolerance <= 0.0:
        return region

    # Merge points of hulls and holes of each polygon
    new_region = pya.Region()
    for poly in region.each():
        new_poly = pya.Polygon(merged_points(list(poly.each_point_hull())))
        for hole_id in range(poly.holes()):
            new_poly.insert_hole(merged_points(list(poly.each_point_hole(hole_id))))
        new_region.insert(new_poly)
    return new_region


def region_with_merged_polygons(region, tolerance, expansion=0.0):
    """ Merges polygons in given region. Ignores gaps that are smaller than given tolerance.

    Arguments:
        region: input region
        tolerance: largest gap size to be ignored
        expansion: the amount by which the polygons are expanded (edges move outwards)

    Returns:
        region with merged polygons
    """
    new_region = region.sized(0.5 * tolerance)  # expand polygons to ignore gaps in merge
    new_region.merge()
    new_region.size(expansion - 0.5 * tolerance)  # shrink polygons back to original shape (+ optional expansion)
    new_region = new_region.smoothed(2)  # smooth out the slight jaggedness on the edges
    return new_region


def match_points_on_edges(cell, layout, layers):
    """ Goes through each polygon edge and splits the edge whenever it passes through a point of another polygon.

    This function can eliminate gaps and overlaps caused by transformation to simple_polygon.

    Arguments:
        cell: A cell object.
        layout: A layout object
        layers: List of layers to be considered and modified
    """
    # Gather points from layers to `all_points` dictionary. This ignores duplicate points.
    all_points = dict()
    for layer in layers:
        shapes = cell.shapes(layout.layer(layer))
        for shape in shapes:
            all_points.update({point: list() for point in shape.simple_polygon.each_point()})
    if not all_points:
        return  # nothing is done if no points exist

    # For each point, assign a list of surrounding points using Voronoi diagram
    point_list = list(all_points)
    vor = spatial.Voronoi([(p.x, p.y) for p in point_list])
    for link in vor.ridge_points:
        all_points[point_list[link[0]]].append(point_list[link[1]])
        all_points[point_list[link[1]]].append(point_list[link[0]])

    # Travel through polygon edges and split edge whenever it passes through a point
    for layer in layers:
        shapes = cell.shapes(layout.layer(layer))
        for shape in shapes:
            points = list(shape.simple_polygon.each_point())
            new_points = []
            for i, p1 in enumerate(points):
                p0 = points[i - 1]
                edge = pya.Edge(p0, p1)
                # Travel from p0 to p1 in Voronoi diagram
                while p0 != p1:
                    # List points that are on the edge towards p1
                    sq_dist = p0.sq_distance(p1)
                    p_on_edge = [p for p in all_points[p0] if edge.contains(p) and p.sq_distance(p1) < sq_dist]
                    if p_on_edge:
                        # Update p0 to be the point on edge towards p1 that is furthest from p1
                        p0 = max(p_on_edge, key=lambda x, y=p1: x.sq_distance(y))
                        new_points.append(p0)  # Add the point to the polygon. Finally, p0 is equal to p1 here.
                    else:
                        # Update p0 to be the neighbour closest to p1
                        p0 = min(all_points[p0], key=lambda x, y=p1: x.sq_distance(y))

            # Replace polygon if any points are added
            if len(new_points) != len(points):
                shapes.replace(shape, pya.SimplePolygon(new_points, True))


def is_clockwise(polygon_points):
    """Returns True if the polygon points are in clockwise order, False if they are counter-clockwise.

    Args:
        polygon_points: list of polygon points, must be either in clockwise or counterclockwise order
    """
    # see https://en.wikipedia.org/wiki/Curve_orientation#Orientation_of_a_simple_polygon
    bottom_left_point_idx = 0
    for idx, point in enumerate(polygon_points[1:]):
        if point.x < polygon_points[bottom_left_point_idx].x and point.y < polygon_points[bottom_left_point_idx].y:
            bottom_left_point_idx = idx
    a = polygon_points[bottom_left_point_idx - 1]
    b = polygon_points[bottom_left_point_idx]
    c = polygon_points[(bottom_left_point_idx + 1) % len(polygon_points)]
    det = (b.x - a.x)*(c.y - a.y) - (c.x - a.x)*(b.y - a.y)
    return det < 0


def circle_polygon(r, n=64, origin=pya.DPoint(0, 0)):
    """
    Returns a polygon for a full circle around the origin.

    Args:
        r: Radius
        origin: Center of the circle, default (0,0)
        n: Number of points.

    Returns: list of ``DPoint``s, length ``n``.
    """
    return pya.DPolygon([origin + pya.DPoint(cos(a / n * 2 * pi) * r, sin(a / n * 2 * pi) * r) for a in range(0, n)])


def arc_points(r, start=0, stop=2 * pi, n=64, origin=pya.DPoint(0, 0)):
    """
    Returns point describing an arc around the origin with specified start and stop angles. The start and stop angle
    are included.

    If start < stop, the points are counter-clockwise; if start > stop, the points are clockwise.

    Args:
        r: Arc radius
        start: Start angle in radians, default 0
        stop: Stop angle in radians, default 2*pi
        origin: Center of the arc, default (0,0)
        n: Number of steps corresponding to a full circle.

    """
    n_steps = max(ceil(n * abs(stop - start) / (2 * pi)), 2)
    step = (stop - start) / (n_steps - 1)
    return [origin + pya.DPoint(cos(start + a * step) * r, sin(start + a * step) * r) for a in range(0, n_steps)]


def _cubic_polynomial(control_points: List[pya.DPoint],
                      spline_matrix: np.array,
                      sample_points: int = 100,
                      endpoint: bool = False) -> List[pya.DPoint]:
    """Returns a list of DPoints sampled uniformly from a third order polynomial spline

    Args:
        control_points: list of exactly four control points
        spline_matrix: matrix of coefficients of the polynomial function
        sample_points: number of points to sample for the curve
        endpoint: if True, will distribute sample points to sample at t = 1.0
    """
    if len(control_points) != 4:
        raise ValueError("There should be exactly four control points for cubic polynomial")
    if spline_matrix.shape != (4, 4):
        raise ValueError("spline_matrix must be of shape (4, 4)")
    geometry_matrix = np.array([[p.x, p.y] for p in control_points]).T
    result_points = []
    for t in np.linspace(0.0, 1.0, sample_points, endpoint=endpoint):
        result_vector = geometry_matrix.dot(spline_matrix).dot(np.array([1, t, t*t, t*t*t]).T)
        result_points.append(pya.DPoint(result_vector[0], result_vector[1]))
    return result_points


def bspline_points(control_points: List[pya.DPoint],
                   sample_points: int = 100,
                   startpoint: bool = False,
                   endpoint: bool = False) -> List[pya.DPoint]:
    """Samples points uniformly from the B-Spline constructed from a sequence of control points.
    The spline is derived from a sequence of cubic splines for each subsequence of four-control points
    in a sliding window.

    Unlike Bezier curves, for each spline in B-Spline it is not guaranteed
    that the first and last control point will be in the spline.

    B-Spline cubic polynomial implemented based on the following reference:
    Kaihuai Qin, "General matrix representations for B-splines", Proceedings Pacific Graphics '98
    Sixth Pacific Conference on Computer Graphics and Applications, Singapore, 1998, pp. 37-43,
    doi: 10.1109/PCCGA.1998.731996

    Args:
        control_points: a sequence of control points, must have at least 4 pya.DPoints elements
        sample_points: number of uniform samples of each cubic B-spline,
            total number of samples is: sample_points * (control_points - 3)
        startpoint: If True, will prepend duplicates of the first control point so that the
            first control point will be in the B-Spline
        endpoint: If True, will append duplicates of the last control point so that the
            last control point will be in the B-Spline

    Returns:
        List of pya.DPoints that can be used as part of a polygon
    """
    # B-Spline doesn't guarantee that the spline will go through the end points,
    # duplicate points on either end if needed
    if startpoint:
        control_points = [control_points[0], control_points[0]] + control_points
    if endpoint:
        control_points = control_points + [control_points[-1], control_points[-1]]
    if len(control_points) < 4:
        raise ValueError("B-Spline needs at least four control points")
    bspline_matrix = (1.0/6.0) * np.array([[ 1,-3, 3,-1],
                                           [ 4, 0,-6, 3],
                                           [ 1, 3, 3,-3],
                                           [ 0, 0, 0, 1]])
    result_points = []
    # Sliding window
    for window_start in range(len(control_points) - 3):
        result_points += _cubic_polynomial( control_points[window_start:window_start+4],
                                            bspline_matrix,
                                            sample_points,
                                            endpoint=(window_start == len(control_points) - 4))
    return result_points


def bezier_points(control_points: List[pya.DPoint], sample_points: int = 100) -> List[pya.DPoint]:
    """Samples points uniformly from the Bezier curve constructed from a sequence of control points.
    The curve is derived from a sequence of cubic splines for each subsequence of four-control points
    such that subsequence shares one control point with the previous subsequence.

    Special care needs to be taken to guarantee continuity in the tangent of the curve.
    The third and fourth control point of each subsequence as well as the second
    control point of the next subsequence have to be in the same line.

    Bezier cubic polynomial implemented based on the following reference:
    Kaihuai Qin, "General matrix representations for B-splines", Proceedings Pacific Graphics '98
    Sixth Pacific Conference on Computer Graphics and Applications, Singapore, 1998, pp. 37-43,
    doi: 10.1109/PCCGA.1998.731996

    Args:
        control_points: a sequence of control points, must be of length equal to 3*n+1 for some integer n
        sample_points: number of uniform samples of each cubic spline,
            total number of samples is: sample_points * ((control_points - 3) / 3)

    Returns:
        List of pya.DPoints that can be used as part of a polygon
    """
    if (len(control_points) - 1) % 3 == 0:
        raise ValueError("For Bezier curve, the number of control points should equal to 3*n+1 for some integer n")
    bezier_matrix = np.array([[ 1,-3, 3,-1],
                              [ 0, 3,-6, 3],
                              [ 0, 0, 3,-3],
                              [ 0, 0, 0, 1]])
    result_points = []
    # Windows with one shared control point
    for window_start in range(0, len(control_points) - 3, 3):
        result_points += _cubic_polynomial( control_points[window_start:window_start+4],
                                            bezier_matrix,
                                            sample_points,
                                            endpoint=(window_start == len(control_points) - 4))
    return result_points



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.elements.element import Element
from kqcircuits.defaults import default_marker_type
import numpy as np


class Marker(Element):
    """Base Class for Markers."""

    default_type = default_marker_type

    diagonal_squares = Param(pdt.TypeInt, "Number of diagonal squares in the marker", 10)
    window = Param(pdt.TypeBoolean, "Window in airbridge flyover layer", False)

    @classmethod
    def create(cls, layout, library=None, marker_type=None, **parameters):
        """Create a Marker cell in layout."""
        return cls.create_subtype(layout, library, marker_type, **parameters)[0]

    def produce_geometry(self):
        """Produce common marker geometry."""

        layer_gap = self.get_layer("base_metal_gap_wo_grid")
        layer_pads = self.get_layer("airbridge_pads")
        layer_flyover = self.get_layer("airbridge_flyover")
        layer_gap_for_ebl = self.get_layer("base_metal_gap_for_EBL")
        layer_protection = self.get_layer("ground_grid_avoidance")

        def insert_to_main_layers(shape):
            self.cell.shapes(layer_gap).insert(shape)
            self.cell.shapes(layer_gap_for_ebl).insert(shape)
            if not self.window:
                self.cell.shapes(layer_flyover).insert(shape)

        # protection for the box
        protection_box = pya.DBox(
            pya.DPoint(220, 220),
            pya.DPoint(-220, -220)
        )
        self.cell.shapes(layer_protection).insert(protection_box)

        # make corners
        corner = pya.DPolygon([
            pya.DPoint(100, 100),
            pya.DPoint(10, 100),
            pya.DPoint(10, 80),
            pya.DPoint(80, 80),
            pya.DPoint(80, 10),
            pya.DPoint(100, 10),
        ])
        inner_corners = [pya.DTrans(a) * corner for a in [0, 1, 2, 3]]
        outer_corners = [pya.DCplxTrans(2, a * 90., False, pya.DVector()) * corner for a in [0, 1, 2, 3]]
        corners = pya.Region([s.to_itype(self.layout.dbu) for s in inner_corners + outer_corners])
        insert_to_main_layers(corners)

        # center box
        sqr_uni = pya.DBox(
            pya.DPoint(10, 10),
            pya.DPoint(-10, -10),
        )
        insert_to_main_layers(sqr_uni)

        self.inv_corners = pya.Region([protection_box.to_itype(self.layout.dbu)])
        self.inv_corners -= corners
        self.cell.shapes(layer_pads).insert(self.inv_corners - pya.Region(sqr_uni.to_itype(self.layout.dbu)))

        # window for airbridge flyover layer
        aflw = pya.DPolygon([
            pya.DPoint(800, 800),
            pya.DPoint(800, 10),
            pya.DPoint(80, 10),
            pya.DPoint(80, 2),
            pya.DPoint(2, 2),
            pya.DPoint(2, 80),
            pya.DPoint(10, 80),
            pya.DPoint(10, 800)
        ])
        if self.window:
            for alpha in [0, 1, 2, 3]:
                self.cell.shapes(layer_flyover).insert(pya.DTrans(alpha) * aflw)

        # marker diagonal
        sqr = pya.DBox(
            pya.DPoint(10, 10),
            pya.DPoint(2, 2),
        )
        self.diagonals = pya.Region()
        for i in range(5, 5 + self.diagonal_squares):
            ds = pya.DCplxTrans(3, 0, False, pya.DVector(50 * i - 3 * 6, 50 * i - 3 * 6)) * sqr
            insert_to_main_layers(ds)
            self.cell.shapes(layer_pads).insert(ds)
            self.diagonals += ds.to_itype(self.layout.dbu)
            self.cell.shapes(layer_protection).insert(
                pya.DCplxTrans(20, 0, False, pya.DVector(50 * i - 20 * 6, 50 * i - 20 * 6)) * sqr)

    @classmethod
    def get_marker_locations(cls, cell_marker, **kwargs):
        """Locations in the wafer for this marker type.
        By default, places four markers at the corners as close as possible
        to the edge clearance.
        Implement this method for your own Marker subclass if you wish to
        have customized placement for your specific marker type.

        Args:
            cls - class that houses this class method
            cell_marker - Marker Cell
            kwargs - keyword arguments needed to determine the mask locations
        Returns:
            A list of placement encoded as DTrans objects that will
            transform the marker cells at their preferred location
        """
        wafer_center_x = kwargs.get('wafer_center_x', 0)
        wafer_center_y = kwargs.get('wafer_center_y', 0)
        wafer_rad = kwargs.get('wafer_rad', 75000)
        edge_clearance = kwargs.get('edge_clearance', 1000)
        margin = kwargs.get('box_margin', 1000)
        _h = cell_marker.dbbox().height()
        _w = cell_marker.dbbox().width()
        coordinate = (wafer_rad - edge_clearance) / np.sqrt(2)
        return [
                pya.DTrans(wafer_center_x - (coordinate - _w/2 - margin), wafer_center_y - (coordinate - _h/2 - margin))
                * pya.DTrans.R180,
                pya.DTrans(wafer_center_x + (coordinate - _w/2 - margin), wafer_center_y - (coordinate - _h/2 - margin))
                * pya.DTrans.R270,
                pya.DTrans(wafer_center_x - (coordinate - _w/2 - margin), wafer_center_y + (coordinate - _h/2 - margin))
                * pya.DTrans.R90,
                pya.DTrans(wafer_center_x + (coordinate - _w/2 - margin), wafer_center_y + (coordinate - _h/2 - margin))
                * pya.DTrans.R0,
                ]

    @classmethod
    def get_marker_region(cls, inst, **kwargs):
        """The Region covered by the marker and surrounding area to be removed from the ground plane.
        By default, a box around the marker extended by the parameter box_margin.
        Implement this method for your own Marker subclass if you wish to
        have a different Region for your specific marker type.

        Args:
            cls - class that houses this class method
            inst - instance of the marker
            kwargs - keyword arguments possibly needed for the region
        Returns:
            pya.Region that can be used to subtract from the ground plane
        """
        margin = kwargs.get('box_margin', 1000)
        return pya.Region(inst.bbox()).extents(margin / inst.cell.layout().dbu)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

from kqcircuits.chips.single_xmons import SingleXmons
from kqcircuits.pya_resolver import pya
from kqcircuits.simulations.port import EdgePort, InternalPort
from kqcircuits.simulations.simulation import Simulation
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.defaults import default_junction_test_pads_type


class SingleXmonsFullChipSim(Simulation):
    n: int
    launchers: bool
    use_test_resonators: bool

    launchers = Param(pdt.TypeBoolean, "True to include launchers in simulation", False)
    use_test_resonators = Param(pdt.TypeBoolean, "True to include XS1-type test resonators. False produces XS2", False)

    def build(self):
        mask_parameters_for_chip = {
            "name_mask": self.name,
            "name_copy": None,
            "with_grid": False,
        }

        chip = self.add_element(SingleXmons, **{
            **mask_parameters_for_chip,
            "name_chip": 'XS1' if self.use_test_resonators else 'XS2',
            "readout_res_lengths": [4490.35, 4578.13, 4668.99, 4763.09, 4860.61, 4961.75],
            "use_test_resonators": self.use_test_resonators,
            "test_res_lengths": [4884.33, 4804.94, 4728.06, 4653.58],
            "n_fingers": 4 * [4],
            "l_fingers": [23.65, 24.204, 24.7634, 25.325],
            "type_coupler": 4 * ["plate"],
            "junction_type": "Sim",
            "n": self.n,
        })

        # Remove unneeded elements
        self.delete_instances(chip, 'Chip Frame')
        self.delete_instances(chip, default_junction_test_pads_type, range(2))

        # Insert chip and get refpoints
        _, refpoints = self.insert_cell(chip, rec_levels=None)

        if not self.launchers:
            # Remove launchers
            self.delete_instances(chip, 'Launcher')

            maximum_box = pya.DBox(pya.DPoint(800, 800), pya.DPoint(9200, 9200))
            port_shift = 0
        else:
            maximum_box = pya.DBox(pya.DPoint(200, 200), pya.DPoint(9800, 9800))
            port_shift = 600

        # Limit the size of the box to fit the ports
        self.box &= maximum_box

        # Define edge ports, shifted inward by port_shift w.r.t. launcher refpoints
        for i, (launcher, shift) in enumerate(zip(
                ['NW', 'WN', 'WS', 'SW', 'SE', 'ES', 'EN', 'NE'],
                [[0, port_shift], [-port_shift, 0], [-port_shift,0], [0, -port_shift],
                 [0, -port_shift], [port_shift, 0], [port_shift, 0], [0, port_shift]])):
            self.ports.append(EdgePort(i + 1, refpoints['{}_port'.format(launcher)] + pya.DVector(*shift)))

        # Add squid internal ports
        for j in range(6):
            self.ports.append(InternalPort(j+9, *self.etched_line(refpoints['qb_{}_port_squid_a'.format(j)],
                                                                  refpoints['qb_{}_port_squid_b'.format(j)])))



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


"""PCell declaration classes for junctions.

Junctions can be either code-generated or loaded from manual design files. They are typically included in
qubits or junction test structures.

This package contains three manually designed SQUIDs in OASIS files from the Quantum Computing and Devices
research group at Aalto:
QCD1, QCD2 and QCD3.
"""

junction_type_choices = [
    'NoSquid',
    'Manhattan',
    'Manhattan Single Junction',
    #'QCD1',
    'Sim',
]



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.elements.element import Element


class FingerCapacitorTaper(Element):
    """The PCell declaration for a tapered finger capacitor.

    Two ports with reference points. Ground plane gap is automatically adjusted to maintain the a/b ratio.

    .. MARKERS_FOR_PNG 2.76,0 10,16 3.6,-30 11.3,0,71.3,0
    """

    finger_number = Param(pdt.TypeInt, "Number of fingers", 5)
    finger_width = Param(pdt.TypeDouble, "Width of a finger", 5, unit="μm")
    finger_gap = Param(pdt.TypeDouble, "Gap between the fingers", 3, unit="μm")
    finger_length = Param(pdt.TypeDouble, "Length of the fingers", 20, unit="μm")
    taper_length = Param(pdt.TypeDouble, "Length of the taper", 60, unit="μm")
    corner_r = Param(pdt.TypeDouble, "Corner radius", 2, unit="μm")

    def can_create_from_shape_impl(self):
        return self.shape.is_path()

    def build(self):
        # shorthand
        n = self.finger_number
        w = self.finger_width
        l = self.finger_length
        g = self.finger_gap
        t = self.taper_length
        total_width = float(n) * (w + g) - g
        a = self.a
        b = self.b

        region_ground = pya.Region([pya.DPolygon([
            pya.DPoint((l + g) / 2, total_width * (b / a) + total_width / 2),
            pya.DPoint((l + g) / 2 + t, b + a / 2),
            pya.DPoint((l + g) / 2 + t, -b - a / 2),
            pya.DPoint((l + g) / 2, -total_width * (b / a) - total_width / 2),
            pya.DPoint(-(l + g) / 2, -total_width * (b / a) - total_width / 2),
            pya.DPoint(-(l + g) / 2 - t, -b - a / 2),
            pya.DPoint(-(l + g) / 2 - t, b + a / 2),
            pya.DPoint(-(l + g) / 2, total_width * (b / a) + total_width / 2),

        ]).to_itype(self.layout.dbu)])

        region_taper_right = pya.Region([pya.DPolygon([
            pya.DPoint((l + g) / 2, total_width / 2),
            pya.DPoint((l + g) / 2 + t, a / 2),
            pya.DPoint((l + g) / 2 + t, -a / 2),
            pya.DPoint((l + g) / 2, -total_width / 2)
        ]).to_itype(self.layout.dbu)])
        region_taper_left = region_taper_right.transformed(pya.Trans().M90)

        polys_fingers = []
        poly_finger = pya.DPolygon([
            pya.DPoint(l / 2, w),
            pya.DPoint(l / 2, 0),
            pya.DPoint(-l / 2, 0),
            pya.DPoint(-l / 2, w)
        ])
        for i in range(0, n):
            trans = pya.DTrans(pya.DVector(g / 2, i * (g + w) - total_width / 2)) if i % 2 else pya.DTrans(
                pya.DVector(-g / 2, i * (g + w) - total_width / 2))
            polys_fingers.append(trans * poly_finger)

        region_fingers = pya.Region([
            poly.to_itype(self.layout.dbu) for poly in polys_fingers
        ])
        region_etch = region_taper_left + region_taper_right + region_fingers
        region_etch.round_corners(self.corner_r / self.layout.dbu, self.corner_r / self.layout.dbu, self.n)

        region_taper_right_small = pya.Region([pya.DPolygon([
            pya.DPoint((l + g) / 2 + self.corner_r, (total_width / 2 - a / 2) * (t - 2 * self.corner_r) / t + a / 2),
            pya.DPoint((l + g) / 2 + t, a / 2),
            pya.DPoint((l + g) / 2 + t, -a / 2),
            pya.DPoint((l + g) / 2 + self.corner_r, -(total_width / 2 - a / 2) * (t - 2 * self.corner_r) / t - a / 2)
        ]).to_itype(self.layout.dbu)])
        region_taper_left_small = region_taper_right_small.transformed(pya.Trans().M90)

        region = (region_ground - region_etch) - region_taper_right_small - region_taper_left_small

        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(region)

        # protection
        region_protection = region_ground.size(0, self.margin / self.layout.dbu, 2)
        self.add_protection(region_protection)

        # ports
        port_a = pya.DPoint(-(l + g) / 2 - t, 0)
        self.add_port("a", port_a, pya.DVector(-1, 0))
        port_b = pya.DPoint((l + g) / 2 + t, 0)
        self.add_port("b", port_b, pya.DVector(1, 0))

        # adds annotation based on refpoints calculated above
